'''Sub-module for tone mapping images.'''
import numpy as np


def make_nonnegative(data):
    ''' Makes an array non-negative by adding the minimum entry to all entries.

    Args:
        data : numpy.ndarray
            Array to make non-negative.

    Returns:
        data_scaled : numpy.ndarray
            Non-negative version of data.
    '''
    minimum = np.nanmin(data)
    if minimum < 0:
        data_scaled = data - np.nanmin(data)
        return data_scaled
    else:
        return data


def normalise(x, vmin=None, vmax=None, clip=True):
    """ Normalise an array of values according to y = (x - vmin / (vmax - vmin).
    
    Args:
        x : numpy.ndarray
            Array of values to normalise.
        vmin : float
            Lower bound for normalisation.
        vmax : float
            Upper bound for normalisation.
        clip : bool
            Clip values to the range [0:1]?
    """
    if not vmin:
        vmin = np.nanmin(x)
    if not vmax:
        vmax = np.nanmax(x)
    y = (x - vmin) / (vmax - vmin)
    if clip:
        y = np.clip(y, 0, 1)
        return y
    else:
        return y


def gamma(data, gamma):
    ''' Apply a gamma correction to the image, raising it to the power 1/gamma.

    Args:
        data : numpy.ndarray
            Data which to gamma correct.

    Returns:
        gamma : float
            The gamma correction factor.
    '''
    return data ** (1 / gamma)



"""Sub-module containing wrappers to LuminosityHDR's HDR tonemapping programs."""
import os
import subprocess
from typing import Optional, Union

import cv2
import numpy
from astropy.io import fits
from PIL import Image
from skimage import io

BASECOMMAND = "luminance-hdr-cli"
if "EVERYSTAMP_LUMINANCE_HDR" in os.environ:
    BASECOMMAND = os.environ["EVERYSTAMP_LUMINANCE_HDR"]


def has_luminance_hdr():
    x = subprocess.run(
        BASECOMMAND,
        shell=True,
        env=dict(os.environ),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    return not x.returncode


def run_command(cmd: Union[list, str]) -> None:
    """Run a command through subprocess.

    Args:
        cmd : list
            List of arguments building up the command or a string of the full command.

    Returns:
        None
    """
    print("Running command:", " ".join(cmd))
    try:
        if isinstance(cmd, list):
            subprocess.run(cmd)
        elif isinstance(cmd, str):
            subprocess.run(cmd.split(" "))
        else:
            raise ValueError("Malformed command: ", cmd)
    except subprocess.CalledProcessError as e:
        print("Tone mapping failed with error code", e.returncode)
        print(e.output)


def _load_tonemapped_tmpdata(name: str, as_gray: bool = True) -> numpy.ndarray:
    """Load the tonemapped output of LuminanceHDR ad return it as an array.

    Args:
        name : str
            File path to the data to load.
        as_gray : bool
            Treat the image as gray scale or not. Default: True.

    Returns:
        data : numpy.ndarray
            NumPy array containing the image.
    """
    # return imageio.imread(name)
    return io.imread(name, as_gray=as_gray)
    # return io.imread(name, as_gray=as_gray)


def _store_tmpfile(data: numpy.ndarray, name: str, header=None) -> str:
    """Store the data to tonemap to a temporary FITS file to pass to LuminanceHDR.

    Args:
        data : numpy.ndarray
            NumPy array containing the image.

    Returns:
        path : str
            Absolute path to the temporary file.
    """
    if name.lower().endswith("fits"):
        print(data.shape)
        fits.writeto(data=data.squeeze(), filename=name, header=header, overwrite=True)
    elif name.lower().endswith("exr"):
        cv2.imwrite(name, data.astype("float32"))
    else:
        im = Image.fromarray(data)
        im.convert(mode="RGB").save(name)
    return os.path.abspath(name)


def ashikhmin(
    data,
    eq2: bool = True,
    simple: Optional[float] = None,
    local_threshold: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Ashikmin 2002.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        eq2 : bool
            Default: True.
        simple : bool
            Default: True.
        local_threshold : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_ashikmin.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo ashikmin "
    cmd += f"--tmoAshEq2 {eq2} "
    cmd += f"--tmoAshSimple {simple} "
    if local_threshold is not None:
        cmd += f"--tmoAshLocal {local_threshold} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def drago(data: numpy.ndarray, bias: float = 0.85) -> numpy.ndarray:
    """Tonemap the image using gradient domain compression as described in Fattal et al. 2002.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        bias : float
            Bias the logarithmic base towards lower or higher values. Default: 0.85.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_drago.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo drago "
    if bias is not None:
        cmd += f"--tmoDrgBias {bias} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    # os.remove(tmpname)
    # os.remove(tmpname_out)
    return data_tm


def duran(
    data: numpy.ndarray,
    sigma_spatial: Optional[float] = None,
    sigma_range: Optional[float] = None,
    base_contrast: Optional[float] = None,
):
    """Tonemap the image using gradient domain compression as described in Durand and Dorsey et al. 2002.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        sigma_spatial : float
            Spatial kernal size. Default: None.
        sigma_range : float
            Range kernel size. Default: None.
        base_contrast : float
            Base contrast. Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_durand.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo durand "
    if sigma_spatial is not None:
        cmd += f"--tmoDurSigmaS {sigma_spatial} "
    if sigma_range is not None:
        cmd += f"--tmoDurSigmaR {sigma_range} "
    if base_contrast is not None:
        cmd += f"--tmoDurBase {base_contrast} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    # os.remove(tmpname)
    # os.remove(tmpname_out)
    return data_tm


def fattal(
    data: numpy.ndarray,
    alpha: Optional[float] = None,
    beta: Optional[float] = None,
    colour_saturation: Optional[float] = None,
    noise: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using gradient domain compression as described in Fattal et al. 2002.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        alpha : float
            Controls which gradient magnitude is preserved when beta is 1. Default is None.
        beta : float
            Controls local detail enhancement.
        colour_saturation : float
            Controls colour saturation.
        noise : float
            Controls the threshold for what is seen as noise, where detail enhancement is reduced.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_fattal.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo fattal "
    if alpha is not None:
        cmd += f"--tmoFatAlpha {alpha} "
    if beta is not None:
        cmd += f"--tmoFatBeta {beta} "
    if colour_saturation is not None:
        cmd += f"--tmoFatAlpha {colour_saturation} "
    if noise is not None:
        cmd += f"--tmoFatAlpha {noise} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def ferradans(
    data: numpy.ndarray, rho: float = -2, inv_alpha: float = 5
) -> numpy.ndarray:
    """Tonemap the image using the method described in Ferradans et al. 2011.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        rho : float
            Controls overall lightness. Larger values yield a brighter image. Default is -2.
        beta : float
            Controls detail enhancement. Larger values yield more detail enhancement. Default is 5.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_ferradans.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo ferradans "
    if rho is not None:
        cmd += f"--tmoFerRho {rho} "
    if inv_alpha is not None:
        cmd += f"--tmoFerInvAlpha {inv_alpha} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def ferwerda(
    data,
    multiplier: Optional[float] = None,
    luminance_adaptation: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using the method described in Ferwerda et al. 1996.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        multiplier : float
            Default is None.
        luminance_adaptation : float
            Default is None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_ferwerda.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo ferradans "
    if multiplier is not None:
        cmd += f"--tmoFerwerdaMul {multiplier} "
    if luminance_adaptation is not None:
        cmd += f"--tmoFerwerdaAdaptLum {luminance_adaptation} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def kimkautz(
    data, c1: Optional[float] = None, c2: Optional[float] = None
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Kim and Kaus 2008.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        c1 : float
            Default is None.
        c2 : float
            Default is None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_kimkautz.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo kimkautz "
    if c1 is not None:
        cmd += f"--tmoKimKautzC1 {c1} "
    if c2 is not None:
        cmd += f"--tmoKimKautzC2 {c2} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def lischinski(data, alpha: Optional[float] = None):
    """Tonemap the image using the method described in Lischinski et al. 2006.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        alpha : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_lischinski.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo lischinski "
    if alpha is not None:
        cmd += f"--tmoLischinskiAlpha {alpha} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def mantiuk06(
    data,
    contrast: Optional[float] = None,
    saturation: Optional[float] = None,
    detail: Optional[float] = None,
    contrast_equalisation: bool = False,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Mantiuk 2006.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        contrast : float
            Default is None.
        saturation : float
            Default is None.
        detail : float
            Default is None.
        contrast_equalisation : bool
            Default is False.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_mantiuk06.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo mantiuk06 "
    if contrast is not None:
        cmd += f"--tmoM06Contrast {contrast} "
    if saturation is not None:
        cmd += f"--tmoM06Saturation {saturation} "
    if detail is not None:
        cmd += f"--tmoM06Detail {detail} "
    cmd += f"--tmoM06ContrastEqual {contrast_equalisation} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def mantiuk08(
    data,
    contrast_enhancement: Optional[float] = None,
    colour_saturation: Optional[float] = None,
    luminance_level: Optional[float] = None,
    set_luminance: bool = False,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Mantiuk 2008.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        contrast_enhancement : float
            Default is None.
        colour_saturation : float
            Default is None.
        luminance_level : float
            Used when set_luminance is True. Default is None.
        set_luminance : bool
            Default is False

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_mantiuk08.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo mantiuk08 "
    if colour_saturation is not None:
        cmd += f"--tmoM08ColorSaturation {colour_saturation} "
    if contrast_enhancement is not None:
        cmd += f"--tmoM08ContrastEnh {contrast_enhancement} "
    if luminance_level is not None:
        cmd += f"--tmoM08LuminanceLvl {luminance_level} "
    cmd += f"--tmoM08SetLuminance {set_luminance} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    # os.remove(tmpname)
    # os.remove(tmpname_out)
    return data_tm


def reinhard02(
    data,
    key: Optional[float] = None,
    phi: Optional[float] = None,
    use_scales: bool = True,
    range: Optional[float] = None,
    low: Optional[float] = None,
    high: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Mantiuk 2008.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        key : float
            Default: None.
        phi : float
            Default: None.
        use_scales : bool
            Default: True.
        range : float
            Default: None.
        low : float
            Default: None.
        high : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_reinhard02.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo reinhard02 "
    if key is not None:
        cmd += f"--tmoR02Key {key} "
    if phi is not None:
        cmd += f"--tmoR02Phi {phi} "
    if use_scales is not None:
        cmd += f"--tmoR02Scales {use_scales} "
    if range is not None:
        cmd += f"--tmoR02Num {range} "
    if low is not None:
        cmd += f"--tmoR02Low {low} "
    if high is not None:
        cmd += f"--tmoR02High {high} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def reinhard05(
    data,
    brightness: Optional[float] = None,
    chroma: Optional[float] = None,
    lightness: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Reinhard 2005.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        brightness : float
            Default: None.
        chroma : float
            Default: None.
        lightness : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_reinhard05.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo reinhard05 "
    if brightness is not None:
        cmd += f"--tmoR05Brightness {brightness} "
    if chroma is not None:
        cmd += f"--tmoR05Chroma {chroma} "
    if lightness is not None:
        cmd += f"--tmoR05Lightness {lightness} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def pattanaik(
    data,
    multiplier: Optional[float] = None,
    local_tonemap: bool = True,
    auto_lum: bool = True,
    cone_level: Optional[float] = None,
    rod_level: Optional[float] = None,
) -> numpy.ndarray:
    """Tonemap the image using the human vision based method described in Pattanaik et al. 2000 and Pattanaik et al. 2002.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        multiplier : float
            Default: None.
        local_tonemap : bool
            Default: True.
        auto_lum : bool
            Default: True.
        cone_level : float
            Default: None.
        rod_level : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_pattanaik.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo pattanaik "
    if multiplier is not None:
        cmd += f"--tmoPatMultiplier {multiplier} "
    if cone_level is not None:
        cmd += f"--tmoPatCone {cone_level} "
    if rod_level is not None:
        cmd += f"--tmoPatCone {rod_level} "
    cmd += f"--tmoPatLocal {local_tonemap} "
    cmd += f"--tmoPatAutoLum {auto_lum} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm


def vanhateren(data, pupil_area: Optional[float] = None):
    """Tonemap the image using the method described in van Hateren 2006.

    Parameters set to None will take their default values as set in LuminanceHDR.

    Args:
        data : numpy.ndarray
            Input data to tonemap.
        pupil_area : float
            Default: None.

    Returns:
        data_tm : numpy.ndarray
            Tonemapped data.
    """
    tmpname = _store_tmpfile(data, "tmp_vanhateren.fits")
    tmpname_out = tmpname.replace(".fits", ".tonemapped.tiff")
    cmd = BASECOMMAND + " -e 0 --tmo vanhateren "
    if pupil_area is not None:
        cmd += f"--tmoVanHaterenPupilArea {pupil_area} "
    cmd += f"-o {tmpname_out} {tmpname}"
    run_command(cmd.split(" "))
    data_tm = _load_tonemapped_tmpdata(tmpname_out)
    os.remove(tmpname)
    os.remove(tmpname_out)
    return data_tm



"""Sub-module for plotting FITS images."""

from typing import List, Optional, Union

import astropy.units as u
import colormaps #noqa F403
import matplotlib.pyplot as plt
import numpy
import aplpy
from aplpy import FITSFigure
from astropy.io import fits
from astropy.visualization import make_lupton_rgb
from astropy.wcs import WCS
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.visualization import (
    ImageNormalize,
    PercentileInterval,
    SqrtStretch,
)
from blend_modes import (
    addition,
    hard_light,
    soft_light,
    lighten_only,
    darken_only,
    multiply,
    dodge,
    difference,
    subtract,
    grain_extract,
    grain_merge,
    divide,
    overlay,
    normal,
)
from matplotlib import colormaps as mplcm
from PIL import Image


def find_rms(image_data: numpy.ndarray) -> float:
    """ Calculate the rms noise of an image. From Cyril Tasse/kMS.

    Args:
        image_data (ndarray): image to calculate the noise of.

    Returns:
        rms (float): rms noise in units of the image.
        
    """
    maskSup = 1e-7
    m = image_data[np.abs(image_data) > maskSup]
    rmsold = np.std(m)
    diff = 1e-1
    cut = 3.0
    med = np.median(m)
    for _ in range(10):
        ind = np.where(np.abs(m - med) < rmsold * cut)[0]
        rms = np.std(m[ind])
        if np.abs((rms - rmsold) / rmsold) < diff:
            break
        rmsold = rms
    print(f"Noise : {str(round(rms * 1000, 4))} {u.mJy/u.beam}")
    return rms


class BasicFITSPlot:
    """Creates a basic plot of a FITS file."""

    def __init__(self, fitsname: str):
        """Initialise a basic plotting object for 2D FITS files.

        Args:
            fitsname : str
                Name of the FITS file that will be plotted.
        """
        self.dpi = 300
        self.figsize = (12, 8)
        self.fitsimage = fitsname
        self.fitsdata = fits.getdata(fitsname).squeeze()
        rgbdata = make_lupton_rgb(
            self.fitsdata[2],
            self.fitsdata[1],
            self.fitsdata[0],
            filename=self.fitsimage.replace(".fits", "_rgb.jpeg"),
            Q=10.0,
            stretch=0.1,
        )
        fits.writeto(
            self.fitsimage.replace(".fits", "_temp_rgb.fits"),
            header=fits.getheader(self.fitsimage),
            data=np.transpose(rgbdata, (2, 0, 1)),
            overwrite=True,
        )
        self.load_rgb = False
        if len(self.fitsdata.shape) > 2:
            # raise NotImplementedError(
            #    "Supplied FITS file appears to have dimensions besides RA and DEC. BasicPlot only supports 2D FITS files."
            # )
            print("Generating RGB image")
            colour_stretch = "linear"
            aplpy.make_rgb_image(
                self.fitsimage.replace(".fits", "_temp_rgb.fits"),
                self.fitsimage.replace(".fits", "_temp_rgb.png"),
                embed_avm_tags=True,
                stretch_r=colour_stretch,
                stretch_g=colour_stretch,
                stretch_b=colour_stretch,
                pmax_r=100,
                pmax_g=100,
                pmax_b=100,
            )
            self.rgbimage = self.fitsimage.replace(".fits", "_temp_rgb.png")
            self.load_rgb = True
        self.data = self.fitsdata
        self.wcs = WCS(fits.getheader(fitsname)).celestial
        self.figsize = [
            self.fitsdata.shape[0] // self.dpi,
            self.fitsdata.shape[1] // self.dpi,
        ]
        if self.figsize[0] < 12:
            self.figsize[0] = 12
        if self.figsize[1] < 8:
            self.figsize[1] = 8

    def get_contour_levels(self, cdata: numpy.ndarray) -> numpy.ndarray:
        """Generate a set of contour levels.

        Args:
            cdata: data to generate the contour levels from.

        Returns:
            contours: levels at which to draw the contours.
        """
        crms = find_rms(cdata)
        contour_levels = (
            np.arange(
                np.sqrt(3 * crms),
                np.sqrt(np.percentile(cdata, 99.999)),
                np.sqrt(2 * np.sqrt(2) * crms),
            )
            ** 2
        )
        print(f"Minimum contour: {3*crms}")
        return contour_levels

    def plot2D(
        self,
        plot_colourbar=False,
        contour_image: Optional[numpy.ndarray] = None,
        contour_levels: Union[int, list] = 7,
        cmap_min: Optional[float] = None,
        cmap_max: Optional[float] = None,
        cmap: str = "grey",
    ):
        """Save a 2D plot of the loaded FITS file.

        Args:
            plot_colourbar : bool
                Add a colour bar to the plot.
            contour_image:
                Add contours based on this image.
            contour_levels:
                Number of contour levels to draw if an integer or contour levels if a list. Defaults to 5.
            cmap_min: lower limit for the colour map.
            cmap_max: upper limit for the colour map.
            cmap: colour map to use for the plotted image.
        """
        if self.load_rgb:
            f = FITSFigure(
                self.fitsimage.replace(".fits", "_temp_rgb.png"), figsize=self.figsize
            )
            f.show_rgb()
            ra = fits.getheader(self.fitsimage)["CRVAL1"]
            dec = fits.getheader(self.fitsimage)["CRVAL2"]
            f.recenter(ra, dec, radius=8.0 / 3600)
        else:
            hdu = fits.PrimaryHDU(
                header=WCS(fits.getheader(self.fitsimage)).celestial.to_header(),
                data=self.data.squeeze(),
            )
            f = FITSFigure(hdu, figsize=self.figsize)
            if not cmap:
                f.show_grayscale(vmin=cmap_min, vmax=cmap_max, pmax=100)
            else:
                print(f"Using colour map: {cmap}")
                f.show_colorscale(vmin=cmap_min, vmax=cmap_max, pmax=100, cmap=cmap)
        if contour_image:
            cdata = fits.getdata(contour_image).squeeze()
            chead = fits.getheader(contour_image)
            contour_levels = self.get_contour_levels(cdata)
            hdu_c = fits.PrimaryHDU(data=cdata, header=chead)
            f.show_contour(hdu_c, levels=contour_levels, colors="white")
        if plot_colourbar:
            f.add_colorbar()
        f.savefig(self.fitsimage.replace("fits", "png"), dpi=self.dpi)

    def plot_noaxes(
        self,
        cmap_min: Optional[float] = None,
        cmap_max: Optional[float] = None,
        cmap: Optional[str] = None,
        contour_image: Optional[str] = None,
    ):
        """Save a plot of the FITS image without any axes.

        Args:
            cmap: colour map to use for plotting the image.
            contour_image: name of the image to take contours from.
            cmap_min: lower limit for the colour map.
            cmap_max: upper limit for the colour map.
        """
        figsize = [
            self.fitsdata.shape[0] // self.dpi,
            self.fitsdata.shape[1] // self.dpi,
        ]
        if figsize[0] < 12:
            figsize[0] = 12
        if figsize[1] < 8:
            figsize[1] = 8
        if self.load_rgb:
            f = FITSFigure("temp_rgb.png", figsize=self.figsize)
            f.show_rgb()
        else:
            hdu = fits.PrimaryHDU(
                header=WCS(fits.getheader(self.fitsimage)).celestial.to_header(),
                data=self.data.squeeze(),
            )
            f = FITSFigure(hdu, figsize=self.figsize)
            if not cmap:
                f.show_grayscale(vmin=cmap_min, vmax=cmap_max, pmax=100)
            else:
                f.show_colorscale(vmin=cmap_min, vmax=cmap_max, pmax=100, cmap=cmap)
        if contour_image:
            cdata = fits.getdata(contour_image).squeeze()
            chead = fits.getheader(contour_image)
            contour_levels = self.get_contour_levels(cdata)
            hdu_c = fits.PrimaryHDU(data=cdata, header=chead)
            f.show_contour(hdu_c, levels=contour_levels, colors="white")
        plt.axis("off")
        f.savefig(
            self.fitsimage.replace("fits", ".noaxes.png"),
            transparent=True,
            dpi=self.dpi,
        )

    def savedata(self, outfile: str):
        """Save data of a BasicPlot object to a FITS file with the same WCS information

        Any tonemapping applied to the original data will be carried over to the FITS file. Physical units will thus be lost.

        Args:
            outfile: name of the output file.
        """
        fits.writeto(
            data=self.data,
            header=self.wcs.to_header(),
            filename=outfile,
            overwrite=True,
        )


class BlendPlot:
    """Creates a composite image using blending modes.

    Attributes:
        backgroundg
        foregroundg
        blend_cmapsg
        centreg
        radiusg
        rmscutg
        blend_modesg
        blend_cmapsg
        blend_opacities

    Methods:
        blend
        load_preset
        prepare_images
        set_blends
    """

    def __init__(
        self,
        background: str,
        foreground: List[str],
        cmaps: List[str],
        centre: SkyCoord,
        radius: float,
        rmscut: float,
    ) -> None:
        """Initialises the BlendPlot.

        Args:
            background (str):
            foreground (str):
            cmaps (list[str]):
            centre (SkyCoord):
            radius (float):
            rmscut (float):
        """
        Image.MAX_IMAGE_PIXELS = None
        self.background = background
        self.foreground = foreground
        self.blend_cmaps = cmaps
        self.centre = centre
        self.radius = radius
        self.rmscut = rmscut

    def prepare_images(self) -> None:
        """Prepare images for blending.

        This involves plotting all the specified images in the same WCS projection, with the specified rms noise cut.
        Plots are saved as PNGs in the current directory.
        """
        print("Preparing background image.")
        fig = FITSFigure(self.background, figsize=(8, 8), dpi=150)
        fig.show_rgb(interpolation="none")
        print(
            f"Recentring on {self.centre.ra.value}, {self.centre.dec.value} {self.radius}"
        )
        fig.recenter(self.centre.ra.value, self.centre.dec.value, self.radius)
        fig.axis_labels.hide()
        fig.tick_labels.hide()
        fig.ticks.hide()
        fig.savefig("temp_background.png")

        for i, fg in enumerate(self.foreground):
            print("Preparing foreground image.")
            h = fits.getheader(fg)
            wcs = WCS(h).celestial
            d = fits.getdata(fg).squeeze()
            rms = find_rms(d)
            d[d < self.rmscut * rms] = np.nan
            norm = ImageNormalize(
                d, interval=PercentileInterval(99.99), stretch=SqrtStretch()
            )

            figf = FITSFigure(self.background, figsize=(8, 8), dpi=150)
            if self.blend_cmaps[i] in list(mplcm):
                cm = self.blend_cmaps[i]
            else:
                cm = eval("colormaps." + self.blend_cmaps[i])
            figf.ax.imshow(
                d,
                transform=figf.ax.get_transform(wcs),
                cmap=cm,
                norm=norm,
                interpolation="none",
            )
            figf.recenter(self.centre.ra.value, self.centre.dec.value, self.radius)
            figf.axis_labels.hide()
            figf.tick_labels.hide()
            figf.ticks.hide()
            figf.savefig(f"temp_foreground_{i:02d}.png", transparent=True)

            fig.ax.imshow(
                d, transform=fig.ax.get_transform(wcs), cmap="afmhot", norm=norm
            )
            fig.savefig(f"temp_reference{i:02d}.png", dpi=150)

        del fig, figf

    def blend(self) -> None:
        """Blend the background and foreground images together using the specified modes.

        Returns:
            None
        """
        img_blend = np.array(Image.open("temp_background.png")).astype(float)
        for i, (bms, bas) in enumerate(zip(self.blend_modes, self.blend_opacities)):
            img_fg = np.array(Image.open(f"temp_foreground_{i:02d}.png")).astype(float)
            if len(bas) == 1:
                opacs = bas * len(bms.split(","))
            elif (len(bas) > 1) and (len(bas) != len(bms.split(","))):
                raise ValueError(
                    "Blend layers and opacities must match in length if nested opacities are given."
                )
            else:
                opacs = bas
            for bm, ba in zip(bms.split(","), opacs):
                print(f"Blending {bm} with {ba} opacity.")
                match bm:
                    case "add":
                        img_blend = addition(img_blend, img_fg, ba)
                    case "softlight":
                        img_blend = soft_light(img_blend, img_fg, ba)
                    case "hardlight":
                        img_blend = hard_light(img_blend, img_fg, ba)
                    case "lighten_only":
                        img_blend = lighten_only(img_blend, img_fg, ba)
                    case "darken_only":
                        img_blend = darken_only(img_blend, img_fg, ba)
                    case "multiply":
                        img_blend = multiply(img_blend, img_fg, ba)
                    case "dodge":
                        img_blend = dodge(img_blend, img_fg, ba)
                    case "difference":
                        img_blend = difference(img_blend, img_fg, ba)
                    case "subtract":
                        img_blend = subtract(img_blend, img_fg, ba)
                    case "grain_extract":
                        img_blend = grain_extract(img_blend, img_fg, ba)
                    case "grain_merge":
                        img_blend = grain_merge(img_blend, img_fg, ba)
                    case "divide":
                        img_blend = divide(img_blend, img_fg, ba)
                    case "overlay":
                        img_blend = overlay(img_blend, img_fg, ba)
                    case "normal":
                        img_blend = normal(img_blend, img_fg, ba)
        Image.fromarray(np.uint8(img_blend)).save(
            self.foreground[0].replace(".fits", "_blend.png")
        )

    def load_preset(self, preset: str) -> None:
        """Loads a preset of blending modes, colour maps and opacities.

        Args:
            preset (str): name of the preset to load.

        Returns:
            None

        Raises:
            ValueError: if an unknown preset is specified.

        """
        match preset:
            case "opt+x-ray+lofar":
                self.blend_modes = ["add,softlight", "add,add,overlay"]
                self.blend_cmaps = ["c_7_16", "solar"]
                self.blend_opacities = [0.35, 0.6]
                self.rmscut = 5.0
            case "opt+lofar_hot":
                self.blend_modes = ["add,add"]
                self.blend_cmaps = ["afmhot"]
                self.blend_opacities = [0.6]
                self.rmscut = 5.0
            case "opt+lofar_solar":
                self.blend_modes = ["add,softlight"]
                self.blend_cmaps = ["solar"]
                self.blend_opacities = [0.6]
                self.rmscut = 5.0
            case _:
                raise ValueError("Unknown preset requested.")

    def set_blends(
        self,
        blend_modes: List[str],
        blend_cmaps: List[str],
        blend_opacities: List[float | List[float]],
    ) -> None:
        """Set the blending modes, colour maps and opacities of the current plot.

        Args:
            blend_modes (list[str]): blending modes to apply to each image.
            blend_cmaps (list[str]): colour maps to use for each image.
            blend_opacities (list[float]): opacities with which to blend each layer.

        Returns:
            None
        """
        self.blend_modes = blend_modes
        self.blend_cmaps = blend_cmaps
        self.blend_opacities = blend_opacities


class SRTPlot:
    """Create a line profile plot similar in style to the SRTPLOT task used by Hogbom 1974."""

    def __init__(self, fitsname: str):
        """Initialise a basic plotting object for 2D FITS files.

        Args:
            fitsname : str
                Name of the FITS file that will be plotted.
        """
        self.dpi = 300
        self.figsize = (12, 12)
        self.fitsimage = fitsname
        self.fitsdata = fits.getdata(fitsname).squeeze()
        if len(self.fitsdata.shape) > 2:
            raise NotImplementedError(
                "Supplied FITS file appears to have dimensions besides RA and DEC. BasicPlot only supports 2D FITS files."
            )
        self.data = self.fitsdata

    def plot2D(self, srt_lines: int = 25, srt_offset: float = 0.01, **kwargs):
        """Save an SRTPLOT style plot of the loaded FITS file.

        Args:
            lines : int
                Number of lines to plot.
            offset : float
                Offset between each line in data units.
        """
        f = plt.figure(figsize=self.figsize)
        ax = f.add_subplot(111)
        stride = self.data.shape[0] // srt_lines
        if stride == 0:
            stride += 1
        for i, line in enumerate(self.data[::stride, ::-1]):
            if i > 0:
                ax.fill_between(
                    list(range(len(line))),
                    (srt_lines - i) * srt_offset,
                    line + (srt_lines - i) * srt_offset,
                    color="w",
                    zorder=i + 2,
                )
            else:
                ax.fill_between(
                    list(range(len(line))),
                    0,
                    line + (srt_lines - i) * srt_offset,
                    color="w",
                    zorder=i + 2,
                )
            ax.plot(line + (srt_lines - i) * srt_offset, color="k", zorder=i + 2)
        ax.set_xlim(0, self.data.shape[1])
        plt.axis("off")
        f.savefig(
            self.fitsimage.replace(".fits", ".srtplot.png"),
            dpi=self.dpi,
            bbox_inches="tight",
            pad_inches=0,
        )


class BasicImagePlot:
    """Creates a basic plot of an image (not FITS) file."""

    def __init__(self, imname: str, wcsimage: Optional[str]=None):
        """Initialise a basic plotting object for 2D FITS files.

        Args:
            fitsname : str
                Name of the FITS file that will be plotted.
        """
        self.dpi = 300
        self.figsize = (12, 8)
        self.image = imname
        import cv2

        self.imdata = cv2.cvtColor(cv2.imread(imname), cv2.COLOR_BGR2RGB)
        self.data = self.imdata
        self.wcsimage = wcsimage
        if wcsimage:
            print(f"Loading WCS from {wcsimage}")
            self.wcs = WCS(fits.getheader(wcsimage)).celestial
        else:
            self.wcs = None

    def plot2D(
        self,
        contour_image: Optional[numpy.ndarray] = None,
        cmap: str = "gray",
        cmap_min: Optional[float] = None,
        cmap_max: Optional[float] = None,
    ):
        """Save a plot of the FITS image without any axes."""
        figsize = [self.imdata.shape[0] // self.dpi, self.imdata.shape[1] // self.dpi]
        if figsize[0] < self.imdata.shape[0]:
            figsize[0] = 8
        if figsize[1] < self.imdata.shape[1]:
            figsize[1] = 8
        print("FIGURE SIZE: ", figsize)
        f = FITSFigure(
            self.wcsimage, figsize=self.figsize, dimensions=[0, 1], slices=[0]
        )
        f.show_rgb(self.image)
        if contour_image:
            cdata = fits.getdata(contour_image).squeeze()
            crms = find_rms(cdata)
            clevels = np.arange(
                5 * crms, np.percentile(cdata, 99.999), np.sqrt(2) * crms
            )
            f.show_contour(contour_image, levels=clevels, colors="w")
        f.savefig(self.image + "_plot.png", dpi=self.dpi)



import subprocess

def run_command(cmd: str) -> int:
    """_summary_

    Parameters
    ----------
    cmd : str
        Command to execute.

    Raises
    ------
    RuntimeError
        If execution failed.

    Returns
    -------
    success : bool
        True if the command exited successfully (return code 0)
    """
    cmd_list = cmd.split(' ')
    output = subprocess.run(cmd_list)
    success = True if output.returncode == 0 else False
    if not success:
        raise RuntimeError('Command "' + cmd + '" failed.')
    return success

def test_help():
    run_command('everystamp -h')

def test_help_download():
    run_command('everystamp download -h')

def test_help_plot():
    run_command('everystamp plot -h')

def test_help_cutout():
    run_command('everystamp cutout -h')


#!/usr/bin/env python
"""Python library aiming to provide a wrapper around various astronomical surveys that offer cutouts."""

__version__ = "1.6.1"
__author__ = "Frits Sweijen"
__license__ = "GPLv3"
import argparse
import logging
import os
import sys
from collections.abc import Iterable
from typing import Generator

import astropy.units as units
import astropy.visualization
import cv2  # type: ignore
import numpy as np
import requests
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astroquery.skyview import SkyView  # type: ignore
from everystamp.cutters import make_cutout_2D, make_cutout_2D_fast, make_cutout_region
from everystamp.tonemapping import lhdr, normalise

logging.basicConfig(
    format="[%(name)s] %(asctime)s - %(levelname)s: %(message)s", level=logging.INFO
)
logger = logging.getLogger("EveryStamp")


# Check if LuminanceHDR is installed.
HAS_LHDR = (
    lhdr.has_luminance_hdr()
)  # True if not subprocess.getstatusoutput('luminance-hdr-cli') else False
# HAS_LHDR = shutil.which('luminance-hdr-cli')


def flatten(xs: Iterable) -> Generator:
    """Generator to flatten a list of nested lists.

    Args:
        xs : list
            The list to flatten.

    Yields:
        x : iterable
            An iterable that will generate the flattened list.
    """
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def _add_args_download(parser):
    """Add arguments to the argparse instance for downloading cutouts.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    #custom_surveys = ["legacy", "pan-starrs", "vlass", "lolss", "lotss", "tgss"]
    try:
        #skyview_surveys = list(flatten(list(SkyView.survey_dict.values())))
        SkyView.survey_dict.values()
    except requests.exceptions.ConnectionError:
        logger.warning(
            "Failed to get SkyView surveys. SkyView cutouts will not be available."
        )
        #skyview_surveys = []
    #allowed_surveys = custom_surveys + skyview_surveys
    required_args = parser.add_argument_group("Required arguments")
    required_args.add_argument(
        "--survey",
        type=str,
        required=True,
        help="Survey from which to download the cutout. Prefix with hips: to download from any of the sky maps listed in the ID column here: http://aladin.cds.unistra.fr/hips/list#hipssky",
    )
    required_args.add_argument(
        "--ra",
        type=float,
        required=False,
        help="Right ascension of cutout centre in degrees. Required if no catalogue is given.",
    )
    required_args.add_argument(
        "--dec",
        type=float,
        required=False,
        help="Declination of cutout centre in degrees. Required if no catalogue is given.",
    )
    required_args.add_argument(
        "--mode",
        type=str,
        required=True,
        default="jpeg",
        choices=["jpeg", "fits", "both"],
        help='Image type to retrieve. Can be "jpeg", "fits" or "both" to retrieve either a JPEG image, FITS file or both. Default value is jpeg.',
    )
    required_args.add_argument(
        "--from_catalogue",
        type=str,
        required=False,
        default="",
        help="Download cutouts from the given catalogue. The catalogue should contain the columns RA and DEC.",
    )

    optional_args = parser.add_argument_group("Optional arguments")
    optional_args.add_argument(
        "--download_dir",
        type=str,
        required=False,
        default=os.getcwd(),
        dest="ddir",
        help="Directory to store downloaded files. If not given will download to $PWD.",
    )
    optional_args.add_argument(
        "--size",
        type=float,
        required=False,
        default=0.01,
        help="Width and height of the square cutout in degrees.",
    )
    optional_args.add_argument(
        "--skyview_pixsize",
        type=float,
        required=False,
        default=1.0,
        help="Pixel size in arcsec for SkyView cutouts.",
    )
    optional_args.add_argument(
        "--hips_pixsize",
        type=float,
        required=False,
        default=1.0,
        help="Pixel size in arcsec for HiPS cutouts.",
    )

    legacy_args = parser.add_argument_group("[DESI Legacy Imaging Surveys]")
    legacy_args.add_argument(
        "--legacy_bands",
        default="grz",
        type=str,
        required=False,
        help="Bands to download. Allowed values are g, r and z. Multiple bands can be specified as a single string. In the case of a JPEG image a colour image will be generated. In the case of a FITS image a FITS cube will be downloaded. Default: grz",
    )
    legacy_args.add_argument(
        "--legacy_layer",
        type=str,
        required=False,
        default="ls-dr9",
        help="Layer to make a cutout from. Default value is ls-dr9. Examples are ls-dr9, sdss or unwise-neo4. See Legacy documentation for all possibilies.",
    )
    legacy_args.add_argument(
        "--legacy_autoscale",
        required=False,
        default=False,
        action="store_true",
        help="Automatically change the pixel size if the resulting image would exceed the server maximum of 3000x3000 pixels.",
    )

    ps_args = parser.add_argument_group("[Pan-STARRS]")
    ps_args.add_argument(
        "--ps_bands",
        type=str,
        required=False,
        default="gri",
        help="Bands to download. Allowed values are g, r and i. Multiple bands can be specified as a single string. Default: gri",
    )

    vlass_args = parser.add_argument_group("[VLASS]")
    vlass_args.add_argument(
        "--vlass_ms",
        type=str,
        required=False,
        default="",
        help="Measurement Set to take the cutout position from.",
    )
    vlass_args.add_argument(
        "--vlass_consider_QA_rejected",
        type=bool,
        required=False,
        default=False,
        help="Also consider tiles that failed the Quality Assurance checks.",
    )
    vlass_args.add_argument(
        "--vlass_type",
        type=str,
        required=False,
        default="ql",
        choices=["ql", "se"],
        help="Image to consider: Quick Look (ql) or Single Epoch (se). Default: ql.",
    )

    lolss_args = parser.add_argument_group("[LoLSS]")
    lolss_args.add_argument(
        "--lolss_release",
        type=str,
        required=False,
        default="pdr",
        choices=["pdr"],
        help="Data release to download from.",
    )

    lotss_args = parser.add_argument_group("[LoTSS]")
    lotss_args.add_argument(
        "--lotss_release",
        type=str,
        required=False,
        default="dr1",
        choices=["pdr", "dr1", "dr2", "dr2-low", "dr3"],
        help="Data release to download from.",
    )


def _add_args_plot(parser):
    """Add arguments to the argparse instance for plotting images.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    required_args = parser.add_argument_group("Required arguments")
    required_args.add_argument(
        "--image", type=str, required=False, help="FITS image to plot."
    )

    required_args = parser.add_argument_group("Optional arguments")
    required_args.add_argument(
        "--style",
        type=str,
        default="normal",
        choices=["normal", "srtplot"],
        required=False,
        help="Style of plot to make.",
    )
    required_args.add_argument(
        "--srt_lines",
        type=int,
        default=25,
        required=False,
        help="Number of lines to draw in an SRTPLOT style plot.",
    )
    required_args.add_argument(
        "--srt_offset",
        type=float,
        default=0.01,
        required=False,
        help="Offset in data units between lines in an SRTPLOT style plot.",
    )
    required_args.add_argument(
        "--gamma",
        type=float,
        default=1.0,
        required=False,
        help="Gamma compress (<1) or expand (>1) an image after tone mapping.",
    )
    required_args.add_argument(
        "--CLAHE",
        action="store_true",
        default=False,
        required=False,
        help="Apply contrast-limited adaptive histogram equalisation.",
    )
    required_args.add_argument(
        "--CLAHE-gridsize",
        default=5,
        type=int,
        required=False,
        help="Grid size to use for CLAHE.",
    )
    required_args.add_argument(
        "--CLAHE-cliplim",
        default=1.0,
        type=float,
        required=False,
        help="Clip limit to use for CLAHE.",
    )
    required_args.add_argument(
        "--stretch",
        default=None,
        type=str,
        required=False,
        choices=["log", "sqrt", "squared", "asinh", "sinh"],
        help="Stretch an image with a certian function.",
    )
    required_args.add_argument(
        "--cmap",
        default="grey",
        type=str,
        required=False,
        help="Colour map to use while plotting.",
    )
    required_args.add_argument(
        "--cmap-min",
        default=None,
        type=float,
        required=False,
        help="Minimum value for the colourmap.",
    )
    required_args.add_argument(
        "--cmap-max",
        default=None,
        type=float,
        required=False,
        help="Maximum value for the colourmap.",
    )

    required_args.add_argument(
        "--contour_image",
        default=None,
        required=False,
        help="Plot the given image as contours over the main image.",
    )
    required_args.add_argument(
        "--wcs_image",
        default=None,
        required=False,
        help="Image to take the WCS from when plotting.",
    )

    if HAS_LHDR:
        required_args.add_argument(
            "--hdr-tonemap",
            default=None,
            type=str,
            choices=[
                "ashikhmin",
                "drago",
                "duran",
                "fattal",
                "ferradans",
                "ferwerda",
                "kimkautz",
                "lischinski",
                "mantiuk06",
                "mantiuk08",
                "pattanaik",
                "reinhard02",
                "reinhard05",
                "vanhateren",
            ],
            required=False,
            help="HDR tonemapping to apply",
        )

        hdr_ashikhmin_args = parser.add_argument_group(
            "HDR Tone mapping -- Ashikhmin et al. 2002 arguments"
        )
        hdr_ashikhmin_args.add_argument(
            "--ashikhmin-eq2",
            default=True,
            type=bool,
            required=False,
            help="Equation 2?",
        )
        hdr_ashikhmin_args.add_argument(
            "--ashikhmin-simple",
            default=True,
            type=bool,
            required=False,
            help="Simple?",
        )
        hdr_ashikhmin_args.add_argument(
            "--ashikhmin-local_threshold",
            default=None,
            type=float,
            required=False,
            help="Local threshold.",
        )

        hdr_drago_args = parser.add_argument_group(
            "HDR Tone mapping -- Drago et al. 2003 arguments"
        )
        hdr_drago_args.add_argument(
            "--drago-bias",
            default=0.85,
            type=float,
            required=False,
            help="Bias parameter controlling the exponent base.",
        )

        hdr_fattal_args = parser.add_argument_group(
            "HDR Tone mapping -- Fattal et al. 2002 arguments"
        )
        hdr_fattal_args.add_argument(
            "--fattal-alpha",
            default=None,
            type=float,
            required=False,
            help="Controls which gradient magnitude is preserved.",
        )
        hdr_fattal_args.add_argument(
            "--fattal-beta",
            default=None,
            type=float,
            required=False,
            help="Controls local detail enhancement.",
        )
        hdr_fattal_args.add_argument(
            "--fattal-colour_saturation",
            default=None,
            type=float,
            required=False,
            help="Controls colour saturation.",
        )
        hdr_fattal_args.add_argument(
            "--fattal-noise",
            default=None,
            type=float,
            required=False,
            help="Controls when local detail enhancement is reduced.",
        )

        hdr_ferradans_args = parser.add_argument_group(
            "HDR Tone mapping -- Ferradans et al. 2011 arguments"
        )
        hdr_ferradans_args.add_argument(
            "--ferradans-rho",
            default=-2,
            type=float,
            required=False,
            help="Controls overall lightness.",
        )
        hdr_ferradans_args.add_argument(
            "--ferradans-inverse_alpha",
            default=5,
            type=float,
            required=False,
            help="Controls local detail enhancement.",
        )

        hdr_ferwerda_args = parser.add_argument_group(
            "HDR Tone mapping -- Ferwerda et al. 1996 arguments"
        )
        hdr_ferwerda_args.add_argument(
            "--ferwerda-multiplier",
            default=None,
            type=float,
            required=False,
            help="Ferwerda multiplier factor.",
        )
        hdr_ferwerda_args.add_argument(
            "--ferwerda-luminance_adaptation",
            default=None,
            type=float,
            required=False,
            help="Ferwerda adaptive luminance factor.",
        )

        hdr_hateren_args = parser.add_argument_group(
            "HDR Tone mapping -- van Hateren 2006 arguments"
        )
        hdr_hateren_args.add_argument(
            "--vanhateren-pupil_area",
            default=None,
            type=float,
            required=False,
            help="Pupil area.",
        )

        hdr_kimkautz_args = parser.add_argument_group(
            "HDR Tone mapping -- Kim and Kautz 2008 arguments"
        )
        hdr_kimkautz_args.add_argument(
            "--kimkautz-c1",
            default=None,
            type=float,
            required=False,
            help="Kim and Kautz c1 factor.",
        )
        hdr_kimkautz_args.add_argument(
            "--kimkautz-c2",
            default=None,
            type=float,
            required=False,
            help="Kim and Kautz c2 factor.",
        )

        hdr_lischinski_args = parser.add_argument_group(
            "HDR Tone mapping -- Lischinski 2006 arguments"
        )
        hdr_lischinski_args.add_argument(
            "--lischinski-alpha",
            default=None,
            type=float,
            required=False,
            help="Alpha.",
        )

        hdr_mantiuk06_args = parser.add_argument_group(
            "HDR Tone mapping -- Mantiuk et al. 2006 arguments"
        )
        hdr_mantiuk06_args.add_argument(
            "--mantiuk06-contrast",
            default=None,
            type=float,
            required=False,
            help="Contrast factor.",
        )
        hdr_mantiuk06_args.add_argument(
            "--mantiuk06-saturation",
            default=None,
            type=float,
            required=False,
            help="Saturation factor.",
        )
        hdr_mantiuk06_args.add_argument(
            "--mantiuk06-detail",
            default=None,
            type=float,
            required=False,
            help="Detail factor.",
        )
        hdr_mantiuk06_args.add_argument(
            "--mantiuk06-contrast_equalisation",
            default=True,
            type=bool,
            required=False,
            help="Equalise contrast?",
        )

        hdr_mantiuk08_args = parser.add_argument_group(
            "HDR Tone mapping -- Mantiuk et al. 2008 arguments"
        )
        hdr_mantiuk08_args.add_argument(
            "--mantiuk08-contrast_enhancement",
            default=None,
            type=float,
            required=False,
            help="Contrast enhancement factor.",
        )
        hdr_mantiuk08_args.add_argument(
            "--mantiuk08-colour_saturation",
            default=None,
            type=float,
            required=False,
            help="Colour saturation factor.",
        )
        hdr_mantiuk08_args.add_argument(
            "--mantiuk08-luminance_level",
            default=0.0,
            type=float,
            required=False,
            help="Luminance level.",
        )
        hdr_mantiuk08_args.add_argument(
            "--mantiuk08-set_luminance",
            default=True,
            type=bool,
            required=False,
            help="Set luminance level?",
        )

        hdr_duran_args = parser.add_argument_group(
            "HDR Tone mapping -- Durand and Dorsey et al. 2002 arguments"
        )
        hdr_duran_args.add_argument(
            "--duran-sigma_spatial",
            default=None,
            type=float,
            required=False,
            help="Spatial kernel size.",
        )
        hdr_duran_args.add_argument(
            "--duran-sigma_range",
            default=None,
            type=float,
            required=False,
            help="Range kernel size.",
        )
        hdr_duran_args.add_argument(
            "--duran-base_contrast",
            default=None,
            type=float,
            required=False,
            help="Base contrast.",
        )

        hdr_reinhard02_args = parser.add_argument_group(
            "HDR Tone mapping -- Reinhard et al. 2002 arguments"
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-key", default=None, type=float, required=False, help="Key."
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-phi", default=None, type=float, required=False, help="Phi."
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-use_scales",
            default=None,
            type=float,
            required=False,
            help="Use scales?",
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-range",
            default=None,
            type=float,
            required=False,
            help="Range.",
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-low", default=None, type=float, required=False, help="Low."
        )
        hdr_reinhard02_args.add_argument(
            "--reinhard02-high", default=None, type=float, required=False, help="High."
        )

        hdr_reinhard05_args = parser.add_argument_group(
            "HDR Tone mapping -- Reinhard et al. 2005 arguments"
        )
        hdr_reinhard05_args.add_argument(
            "--reinhard05-brightness",
            default=None,
            type=float,
            required=False,
            help="Brightness.",
        )
        hdr_reinhard05_args.add_argument(
            "--reinhard05-chroma",
            default=None,
            type=float,
            required=False,
            help="Chroma.",
        )
        hdr_reinhard05_args.add_argument(
            "--reinhard05-lightness",
            default=None,
            type=float,
            required=False,
            help="Lightness.",
        )

        hdr_pattanaik_args = parser.add_argument_group(
            "HDR Tone mapping -- Pattanaik arguments"
        )
        hdr_pattanaik_args.add_argument(
            "--pattanaik-multiplier",
            default=None,
            type=float,
            required=False,
            help="Multiplier.",
        )
        hdr_pattanaik_args.add_argument(
            "--pattanaik-local_tonemap",
            default=True,
            type=bool,
            required=False,
            help="Multiplier.",
        )
        hdr_pattanaik_args.add_argument(
            "--pattanaik-auto_lum",
            default=True,
            type=bool,
            required=False,
            help="Automatic luminance?",
        )
        hdr_pattanaik_args.add_argument(
            "--pattanaik-cone_level",
            default=None,
            type=float,
            required=False,
            help="Cone level.",
        )
        hdr_pattanaik_args.add_argument(
            "--pattanaik-rod_level",
            default=None,
            type=float,
            required=False,
            help="Rod level.",
        )
    else:
        logger.warning(
            "Cannot find luminance-hdr-cli. HDR tone mapping functionality will not be available unless LuminanceHDR is (correctly) installed."
        )


def _add_args_cutout(parser):
    """Add arguments to the argparse instance for making cutouts.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    required_args = parser.add_argument_group("Required arguments")
    required_args.add_argument(
        "--image", type=str, required=False, help="FITS image to make a cutout from."
    )
    required_args.add_argument(
        "--ra",
        type=float,
        required=True,
        help="Right ascension of cutout centre in degrees.",
    )
    required_args.add_argument(
        "--dec",
        type=float,
        required=True,
        help="Declination of cutout centre in degrees.",
    )
    required_args.add_argument(
        "--size",
        type=float,
        required=False,
        default=0.01,
        help="Width and height of the square cutout in degrees.",
    )
    required_args.add_argument(
        "--region",
        type=str,
        required=False,
        default="",
        help="DS9 region file to trim the image to.",
    )
    required_args.add_argument(
        "--from_catalogue",
        type=str,
        required=False,
        default="",
        help="Download cutouts from the given catalogue. The catalogue should contain the columns RA and DEC.",
    )
    required_args.add_argument(
        "--cutout-mode",
        type=str,
        required=False,
        default="fast",
        choices=["fast", "astropy"],
        help="Affects the way cutouts are made. `astropy` uses astropy's Cutout2D, while `fast` simply slices the array. Does not affect cutouts made with a region file.",
    )

    optional_args = parser.add_argument_group("Optional arguments")
    optional_args.add_argument(
        "--output-dir",
        type=str,
        required=False,
        default=os.getcwd(),
        dest="ddir",
        help="Directory to store cutout files. If not given will saved to $PWD.",
    )


def _add_args_composite(parser):
    """Add arguments to the argparse instance for making composites.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    required_args = parser.add_argument_group("Required arguments")
    required_args.add_argument(
        "--background", type=str, required=False, help="Background image."
    )
    required_args.add_argument(
        "--foreground", type=str, nargs="+", required=False, help="Foreground image."
    )
    required_args.add_argument(
        "--radius",
        type=float,
        required=False,
        help="Radius around the centre of the foreground image to consider.",
    )
    required_args.add_argument(
        "--blend-modes",
        type=str,
        nargs="+",
        required=False,
        help="Blending mode to blend the foreground image into the background image with. Available modes are: soft_light, lighten_only, dodge, add, darken_only, multiply, hard_light, difference, subtract, grain-extract, grain_merge, divide, overlay and normal. Multiple modes can be applied per layer using , as an in-layer separator.",
    )
    required_args.add_argument(
        "--blend-opacities",
        type=str,
        nargs="+",
        required=False,
        help="Opacity to blend the foreground image into the background image with. If multiple blend modes are given per layer, in-layer opacities for each mode can be given by separating them with ,.",
    )
    required_args.add_argument(
        "--blend-cmaps",
        type=str,
        nargs="+",
        required=False,
        help="Colour maps to use for each blending layer.",
    )

    optional_args = parser.add_argument_group("Optional arguments")
    optional_args.add_argument(
        "--bg_wcs_from",
        type=str,
        default=None,
        help="File to extract the background WCS from. Can be a text file with a FITS header, or a FITS image. If the background image has AVM data or is a FITS image this is not needed.",
    )
    optional_args.add_argument(
        "--fg_rms_cut",
        type=float,
        default=np.nan,
        help="Factor of the rms noise level below which to blank the foreground image. Mainly useful for radio overlay.",
    )

    optional_args.add_argument(
        "--preset",
        type=str,
        default="",
        help="Apply a preset of blending modes.",
        choices=["opt+lofar_hot", "opt+lofar_solar", "opt+x-ray+lofar"],
    )


def _process_args_download(args):
    """Process arguments to the argparse instance for downloading cutouts.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    if not (args.ra and args.dec) and (not args.from_catalogue):
        raise ValueError(
            "Either right ascension and declination or a catalogue must be specified to download from."
        )
    elif (args.ra and args.dec) and (args.from_catalogue):
        raise ValueError(
            "Cannot specify both right ascension and declination and a catalogue."
        )
    elif ((args.ra and not args.dec) or (not args.ra and args.dec)) and (
        args.from_catalogue
    ):
        raise ValueError("Need to specify both right ascension and declination.")
    ras = []
    decs = []
    if (args.ra and args.dec) and (not args.from_catalogue):
        ras.append(args.ra)
        decs.append(args.dec)
    elif args.from_catalogue:
        tab = Table.read(args.from_catalogue)
        ras = tab["RA"]
        decs = tab["DEC"]
    if args.ddir and (not os.path.exists(args.ddir)):
        logger.info("Download directory does not exist, creating it")
        os.mkdir(args.ddir)
    logger.info("Survey is %s", args.survey)
    if args.lotss_release == "dr3":
        print("LoTSS DR3 is not yet publicly available. Please authenticate.")
        import getpass
        user = getpass.getpass("username:")
        password = getpass.getpass("password:")
    for ra, dec in zip(ras, decs):
        if args.survey.startswith("hips:"):
            from everystamp.downloaders import HiPSDownloader

            if "http" in args.survey:
                hipsname = args.survey[5:]
            else:
                hipsname = args.survey.split(":")[1]
            vd = HiPSDownloader(hips=hipsname, name=hipsname.replace("/", "_"))
            vd.download(
                ra=ra,
                dec=dec,
                size=args.size,
                pixsize=args.hips_pixsize,
                ddir=args.ddir,
                mode=args.mode.replace("e", ""),
            )
        elif args.survey == "legacy":
            from everystamp.downloaders import LegacyDownloader

            ld = LegacyDownloader()
            ld.download(
                ra=ra,
                dec=dec,
                bands=args.legacy_bands,
                mode=args.mode,
                size=args.size,
                layer=args.legacy_layer,
                autoscale=args.legacy_autoscale,
                ddir=args.ddir,
            )
        elif args.survey == "pan-starrs":
            from everystamp.downloaders import PanSTARRSDownloader

            pd = PanSTARRSDownloader()
            pd.download(
                ra=ra,
                dec=dec,
                bands=args.ps_bands,
                mode=args.mode,
                size=args.size,
                ddir=args.ddir,
            )
        elif args.survey == "first":
            if args.mode == "both":
                raise ValueError("FIRST download does not support `both` (yet).")
            from everystamp.downloaders import FIRSTDownloader
            vd = FIRSTDownloader()
            vd.download(
                ra=ra,
                dec=dec,
                size=args.size,
                mode=args.mode,
                ddir=args.ddir,
            )
        elif args.survey == "vlass":
            if args.mode == "both" or args.mode == "jpeg":
                raise ValueError("VLASS download does not support JPEG (yet).")
            from everystamp.downloaders import VLASSDownloader

            vd = VLASSDownloader(datatype=args.vlass_type)
            vd.download(
                ra=ra,
                dec=dec,
                size=args.size,
                crop=True,
                consider_QA_rejected=args.vlass_consider_QA_rejected,
                ddir=args.ddir,
            )
        elif args.survey == "lolss":
            if args.mode == "both" or args.mode == "jpeg":
                raise ValueError("LoLLS download does not support JPEG (yet).")
            from everystamp.downloaders import VODownloader

            vd = VODownloader(
                url="https://vo.astron.nl/lolss/q/cutout/siap.xml", name="LoLSS"
            )
            vd.download(ra=ra, dec=dec, size=args.size, ddir=args.ddir)
        elif args.survey == "lotss":
            if (args.mode == "both" or args.mode == "jpeg") and (
                args.lotss_release != "dr2"
            ):
                raise ValueError(
                    "LoTSS {:s} download does not support JPEG (yet).".format(
                        args.lotss_release.upper()
                    )
                )
            from everystamp.downloaders import VODownloader

            if args.lotss_release == "pdr":
                vd = VODownloader(
                    url="https://vo.astron.nl/lofartier1/q_img/cutout/siap.xml",
                    name="LoTSS-PDR",
                )
            elif args.lotss_release == "dr1":
                vd = VODownloader(
                    url="https://vo.astron.nl/hetdex/lotss-dr1-img/cutout/siap.xml",
                    name="LoTSS-DR1",
                )
                vd.download(ra=ra, dec=dec, size=args.size, ddir=args.ddir)
            elif args.lotss_release == "dr2" or args.lotss_release == "dr2-low":
                from everystamp.downloaders import LoTSSDownloader

                vd = LoTSSDownloader()
                vd.download(
                    ra=ra,
                    dec=dec,
                    size=args.size,
                    ddir=args.ddir,
                    mode=args.mode.replace("e", ""),
                    release=args.lotss_release,
                )
            elif args.lotss_release == "dr3":
                from everystamp.downloaders import LoTSSDownloader

                vd = LoTSSDownloader()
                vd.download(
                    ra=ra,
                    dec=dec,
                    size=args.size,
                    ddir=args.ddir,
                    mode=args.mode.replace("e", ""),
                    release=args.lotss_release,
                    credentials=(user, password),
                )
        elif args.survey == "tgss":
            if args.mode == "both" or args.mode == "jpeg":
                raise ValueError("TGSS download does not support JPEG (yet).")
            from everystamp.downloaders import VODownloader

            vd = VODownloader(
                url="https://vo.astron.nl/tgssadr/q_fits/cutout/siap.xml", name="TGSS"
            )
            vd.download(ra=ra, dec=dec, size=args.size, ddir=args.ddir)
        else:
            if args.mode == "both" or args.mode == "jpeg":
                raise ValueError("SkyView download does not support JPEG (yet).")
            from everystamp.downloaders import SkyViewDownloader

            sd = SkyViewDownloader(args.survey)
            sd.download(
                ra=ra,
                dec=dec,
                size=args.size,
                pixsize=args.skyview_pixsize,
                ddir=args.ddir,
            )


def _process_args_plot(args):
    """Process arguments to the argparse instance for plotting images.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    logger.info("Plotting image %s", args.image)
    import numpy as np

    from everystamp.plotters import BasicFITSPlot, BasicImagePlot, SRTPlot
    from everystamp.tonemapping import gamma
    from everystamp.tonemapping.lhdr import (
        ashikhmin,
        drago,
        duran,
        fattal,
        ferradans,
        ferwerda,
        kimkautz,
        lischinski,
        mantiuk06,
        mantiuk08,
        pattanaik,
        reinhard02,
        reinhard05,
        vanhateren,
    )

    if args.image.lower().endswith("fits"):
        if args.style == "normal":
            bp = BasicFITSPlot(args.image)
        elif args.style == "srtplot":
            bp = SRTPlot(args.image)
    else:
        # Probably an image format.
        bp = BasicImagePlot(args.image, wcsimage=args.wcs_image)
    if HAS_LHDR and args.hdr_tonemap:
        bp.data = normalise(bp.data)
        if args.hdr_tonemap == "ashikhmin":
            logger.info("Tonemapping image with ashikhmin")
            bp.data = ashikhmin(
                bp.data,
                eq2=args.ashikhmin_eq2,
                simple=args.ashikhmin_simple,
                local_threshold=args.ashikhmin_local_threshold,
            )
        if args.hdr_tonemap == "fattal":
            logger.info("Tonemapping image with fattal")
            bp.data = fattal(
                bp.data,
                alpha=args.fattal_alpha,
                beta=args.fattal_beta,
                colour_saturation=args.fattal_colour_saturation,
                noise=args.fattal_noise,
            )
        if args.hdr_tonemap == "drago":
            logger.info("Tonemapping image with drago")
            bp.data = drago(bp.data, bias=args.drago_bias)
        if args.hdr_tonemap == "ferradans":
            logger.info("Tonemapping image with ferradans")
            bp.data = ferradans(
                bp.data, rho=args.ferradans_rho, inv_alpha=args.ferradans_inverse_alpha
            )
        if args.hdr_tonemap == "ferwerda":
            logger.info("Tonemapping image with ferwerda")
            bp.data = ferwerda(
                bp.data,
                multiplier=args.ferwerda_multiplier,
                luminance_adaptation=args.ferwerda_luminance_adaptation,
            )
        if args.hdr_tonemap == "kimkautz":
            logger.info("Tonemapping image with kimkautz")
            bp.data = kimkautz(bp.data, c1=args.kimkautz_c1, c2=args.kimkautz_c2)
        if args.hdr_tonemap == "lischinski":
            logger.info("Tonemapping with lischinski")
            bp.data = lischinski(bp.data, alpha=args.lischinski_alpha)
        if args.hdr_tonemap == "mantiuk06":
            logger.info("Tonemapping image with mantiuk06")
            bp.data = mantiuk06(
                bp.data,
                contrast=args.mantiuk06_contrast,
                saturation=args.mantiuk06_saturation,
                detail=args.mantiuk06_detail,
                contrast_equalisation=args.mantiuk06_contrast_equalisation,
            )
        if args.hdr_tonemap == "mantiuk08":
            logger.info("Tonemapping image with mantiuk08")
            bp.data = mantiuk08(
                bp.data,
                contrast_enhancement=args.mantiuk08_colour_saturation,
                colour_saturation=args.mantiuk08_colour_saturation,
                luminance_level=args.mantiuk08_luminance_level,
                set_luminance=args.mantiuk08_set_luminance,
            )
        if args.hdr_tonemap == "duran":
            logger.info("Tonemapping image with duran")
            bp.data = duran(
                bp.data,
                sigma_spatial=args.duran_sigma_spatial,
                sigma_range=args.duran_sigma_range,
                base_contrast=args.duran_base_contrast,
            )
        if args.hdr_tonemap == "pattanaik":
            logger.info("Tonemapping image with pattanaik")
            bp.data = pattanaik(
                bp.data,
                multiplier=args.pattanaik_multiplier,
                local_tonemap=args.pattanaik_local_tonemap,
                auto_lum=args.pattanaik_auto_lum,
                cone_level=args.pattanaik_cone_level,
                rod_level=args.pattanaik_rod_level,
            )
        if args.hdr_tonemap == "reinhard02":
            logger.info("Tonemapping image with reinhard02")
            bp.data = reinhard02(
                bp.data,
                key=args.reinhard02_key,
                phi=args.reinhard02_phi,
                use_scales=args.reinhard02_use_scales,
                range=args.reinhard02_range,
                low=args.reinhard02_low,
                high=args.reinhard02_high,
            )
        if args.hdr_tonemap == "reinhard05":
            logger.info("Tonemapping image with reinhard05")
            bp.data = reinhard05(
                bp.data,
                brightness=args.reinhard05_brightness,
                chroma=args.reinhard05_chroma,
                lightness=args.reinhard05_lightness,
            )
        if args.hdr_tonemap == "vanhateren":
            logger.info("Tonemapping with vanhateren")
            bp.data = vanhateren(bp.data, pupil_area=args.vanhateren_pupil_area)

    if args.CLAHE:
        if args.image.lower().endswith("fits"):
            # bp.data = make_nonnegative(bp.fitsdata)
            # bp.data /= np.nanmax(bp.data)
            bp.data = normalise(bp.fitsdata)
            bp.data *= 2**16
            bp.data = bp.data.astype(np.uint16)
            clahe = cv2.createCLAHE(
                clipLimit=args.CLAHE_cliplim,
                tileGridSize=(args.CLAHE_gridsize, args.CLAHE_gridsize),
            )
            bp.data = clahe.apply(bp.data)
        else:
            clahe = cv2.createCLAHE(
                clipLimit=args.CLAHE_cliplim,
                tileGridSize=(args.CLAHE_gridsize, args.CLAHE_gridsize),
            )
            Lab = cv2.cvtColor(bp.data, cv2.COLOR_RGB2Lab)
            L, a, b = cv2.split(Lab)
            L_clahe = clahe.apply(L)
            Lab_clahe = cv2.merge((L_clahe, a, b))
            cv2.cvtColor(Lab_clahe, cv2.COLOR_Lab2RGB)

    if args.gamma != 1:
        logger.info("Applying gamma stretch of {:f}".format(args.gamma))
        if args.image.lower().endswith("fits"):
            bp.data = gamma(normalise(bp.data), args.gamma)
        else:
            Lab = cv2.cvtColor(bp.data.astype(np.uint8), cv2.COLOR_RGB2Lab)
            L, a, b = cv2.split(Lab)
            lookUpTable = np.empty((1, 256), np.uint8)
            for i in range(256):
                lookUpTable[0, i] = np.clip(pow(i / 255.0, args.gamma) * 255.0, 0, 255)
            L_gamma = cv2.LUT(L, lookUpTable)
            Lab_gamma = cv2.merge((L_gamma, a, b))
            bp.data = cv2.cvtColor(Lab_gamma, cv2.COLOR_Lab2RGB)

    if args.stretch:
        # 'log10', 'sqrt', 'squared', 'asinh', 'sinh'
        if args.stretch == "log":
            stretch = astropy.visualization.LogStretch()
        elif args.stretch == "sqrt":
            stretch = astropy.visualization.SqrtStretch()
        elif args.stretch == "squared":
            stretch = astropy.visualization.SquaredStretch()
        elif args.stretch == "asinh":
            stretch = astropy.visualization.AsinhStretch()
        elif args.stretch == "sinh":
            stretch = astropy.visualization.SinhStretch()
        bp.data = stretch(normalise(bp.data), min)

    if args.contour_image and (args.style == "normal"):
        bp.plot2D(
            contour_image=args.contour_image,
            cmap=args.cmap,
            cmap_min=args.cmap_min,
            cmap_max=args.cmap_max,
        )
    elif (not args.contour_image) and (args.style == "normal"):
        bp.plot2D(cmap=args.cmap, cmap_min=args.cmap_min, cmap_max=args.cmap_max)
    elif args.style == "srtplot":
        bp.plot2D(srt_lines=args.srt_lines, srt_offset=args.srt_offset)
    if args.image.lower().endswith("fits") and (args.style == "normal"):
        bp.savedata(args.image.replace(".fits", ".tonemapped.fits"))
        if args.contour_image:
            bp.plot_noaxes(cmap=args.cmap, cmap_min=args.cmap_min, cmap_max=args.cmap_max, contour_image=args.contour_image)
        else:
            bp.plot_noaxes(cmap=args.cmap, cmap_min=args.cmap_min, cmap_max=args.cmap_max)


def _process_args_cutout(args):
    """Process arguments to the argparse instance for downloading cutouts.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    if args.ddir and (not os.path.exists(args.ddir)):
        logger.info("Download directory does not exist, creating it")
        os.mkdir(args.ddir)
    s = args.size * units.deg

    if (args.ra and args.dec) and (not args.from_catalogue):
        ras = [args.ra]
        decs = [args.dec]
    elif args.from_catalogue:
        tab = Table.read(args.from_catalogue)
        ras = tab["RA"]
        decs = tab["DEC"]
    coords = SkyCoord(ras, decs, unit="deg")

    if args.region and (args.from_catalogue or args.size):
        raise ValueError("Cannot specify a region, and a catalogue or size at the same time.")
    if args.region and (args.cutout_mode == "fast"):
        raise NotImplementedError("Cutout mode `fast` is not supported with regions.")

    if args.region:
        for c in coords:
            out = os.path.join(
                args.ddir,
                os.path.basename(args.image).replace(
                    ".fits",
                    ".cropped_{:.4f}_{:.4f}_{:.4f}.fits".format(
                        c.ra.value, c.dec.value, args.size
                    ),
                ),
            )
            make_cutout_region(args.image, region=args.region, outfile=out)
    else:
        if args.cutout_mode == "fast":
            cutout_func = make_cutout_2D_fast
        elif args.cutout_mode == "astropy":
            cutout_func = make_cutout_2D
        else:
            raise ValueError(f"Invalid cutout mode {args.cutout_mode} encountered.")
        for c in coords:
            out = os.path.join(
                args.ddir,
                os.path.basename(args.image).replace(
                    ".fits",
                    ".cropped_{:.4f}_{:.4f}_{:.4f}.fits".format(
                        c.ra.value, c.dec.value, args.size
                    ),
                ),
            )
            cutout_func(args.image, pos=c, size=s, outfile=out)


def _process_args_composite(args):
    """Process arguments to the argparse instance for making composites.

    Args:
        parser : ArgumentParser
            ArgumentParser instance to which to add entries.
    """
    from astropy.io import fits
    from everystamp.plotters import BlendPlot
    import magic
    import pyavm

    if args.bg_wcs_from:
        filetype = magic.from_file(args.bg_wcs_from).split()
        if "FITS" in filetype:
            print(f"Extracting WCS information from FITS file {args.bg_wcs_from}.")
            try:
                header = fits.getheader(args.bg_wcs_from)
            except OSError:
                with open(args.bg_wcs_from, "rb") as f:
                    header = fits.Header.fromfile(
                        f, sep="\n", padding=False, endcard=False
                    )
            avm = pyavm.AVM.from_header(header)
            avm.embed(args.background, os.path.basename(args.background) + ".avm.png")
        elif "ASCII" in filetype:
            print(f"Extracting WCS information from ASCII file {args.bg_wcs_from}.")
            raise NotImplementedError
        else:
            print(f"Could not parse WCS from {args.bg_wcs_from}; unknown file type.")
            sys.exit(-1)
    else:
        try:
            avm = pyavm.AVM.from_image(args.background)
            print(avm.to_wcs())
        except pyavm.exceptions.NoXMPPacketFound:
            if args.bg_wcs_from:
                print("No AVM metadata found in background image.")
                filetype = magic.from_file(args.bg_wcs_from).split()
                if "FITS" in filetype:
                    print(
                        f"Extracting WCS information from FITS file {args.bg_wcs_from}."
                    )
                    try:
                        header = fits.getheader(args.bg_wcs_from)
                    except OSError:
                        with open(args.bg_wcs_from, "rb") as f:
                            header = fits.Header.fromfile(
                                f, sep="\n", padding=False, endcard=False
                            )
                    avm = pyavm.AVM.from_header(header)
                    avm.embed(
                        args.background,
                        os.path.basename(args.background) + ".avm.png",
                    )
                elif "ASCII" in filetype:
                    print(
                        f"Extracting WCS information from ASCII file {args.bg_wcs_from}."
                    )
                    raise NotImplementedError
                else:
                    print(
                        f"Could not parse WCS from {args.bg_wcs_from}; unknown file type."
                    )
                    sys.exit(-1)
            else:
                print(
                    "Cannot make composite: No AVM metadata found in background image and no WCS file given."
                )
                sys.exit(-1)

    background_file = os.path.basename(args.background) + ".avm.png"
    header_fg = fits.getheader(args.foreground[0])
    pos = SkyCoord(header_fg["CRVAL1"], header_fg["CRVAL2"], unit="deg")
    bp = BlendPlot(
        background_file,
        args.foreground,
        args.blend_cmaps,
        pos,
        args.radius,
        args.fg_rms_cut,
    )
    if args.preset:
        bp.load_preset(args.preset)
    else:
        opacities = [
            [float(o) for o in layer.split(",")] for layer in args.blend_opacities
        ]
        print(opacities)
        bp.set_blends(args.blend_modes, args.blend_cmaps, opacities)
    bp.prepare_images()
    bp.blend()


def main():
    """Main entry point if called as a standalone executable."""
    parser = argparse.ArgumentParser(
        description="EveryStamp {:s} by {:s}".format(__version__, __author__)
    )
    parser._action_groups.pop()

    subparsers = parser.add_subparsers(
        dest="cmd", description="Description of sub commands."
    )
    subparser_dl = subparsers.add_parser(
        "download",
        description="Download a cutout from a user-specified survey. See everystamp download -h for all available parameters.",
        help="Download a cutout from a specified survey.",
    )
    _add_args_download(subparser_dl)

    subparser_plot = subparsers.add_parser(
        "plot",
        description="Plot a given FITS image. See everystamp plot -h for more information.",
        help="Plot a user-supplied FITS image.",
    )
    _add_args_plot(subparser_plot)

    subparser_cut = subparsers.add_parser(
        "cutout",
        description="Trim a given FITS image. See everystamp cutout -h for more information.",
        help="Cut a user-supplied FITS image to size.",
    )
    _add_args_cutout(subparser_cut)

    subparser_composite = subparsers.add_parser(
        "composite",
        description="Create a composite of a background image and foreground image(s).",
        help="Compose background and foreground image(s) with various blending modes.",
    )
    _add_args_composite(subparser_composite)

    args = parser.parse_args()

    if args.cmd == "download":
        _process_args_download(args)
    if args.cmd == "plot":
        _process_args_plot(args)
    if args.cmd == "cutout":
        _process_args_cutout(args)
    if args.cmd == "composite":
        _process_args_composite(args)


if __name__ == "__main__":
    main()



from test_basics import run_command

def test_legacy_cli():
    run_command('everystamp download --survey legacy --ra 165.345 --dec 65.567 --size 0.1 --mode jpeg --legacy_bands grz --legacy_layer ls-dr9 --legacy_autoscale')
    run_command('everystamp download --survey legacy --ra 202.4841667 --dec 47.2305556 --size 0.1 --mode fits --legacy_bands z --legacy_layer ls-dr9 --legacy_autoscale')

def test_LoLSS_cli():
    run_command('everystamp download --survey lolss --ra 202.4841667 --dec 47.2305556 --size 0.1 --mode fits --download_dir tests_output')

def test_LoTSS_cli():
    run_command('everystamp download --survey lotss --ra 202.4841667 --dec 47.2305556 --size 0.3 --mode fits --download_dir tests_output --lotss_release pdr')
    run_command('everystamp download --survey lotss --ra 202.4841667 --dec 47.2305556 --size 0.3 --mode fits --download_dir tests_output --lotss_release dr1')
    run_command('everystamp download --survey lotss --ra 202.4841667 --dec 47.2305556 --size 0.3 --mode fits --download_dir tests_output --lotss_release dr2')

def test_PanSTARRS_cli():
    run_command('everystamp download --survey pan-starrs --ra 165.345 --dec 65.567 --size 0.1 --mode jpeg --ps_bands gri --download_dir tests_output')

def test_TGSS_cli():
    run_command('everystamp download --survey tgss --ra 165.345 --dec 65.567 --size 0.1 --download_dir tests_output --mode fits')

def test_VLASS_cli():
    run_command('everystamp download --survey vlass --ra 165.345 --dec 65.567 --size 0.1 --download_dir tests_output --mode fits --vlass_type ql')
    #run_command('everystamp download --survey vlass --ra 180.345 --dec 60.567 --size 0.1 --download_dir tests_output --mode fits --vlass_type se')



"""Sub-module for trimming FITS images."""

import numpy as np
import pyregion

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.io.fits import Header
from astropy.nddata import Cutout2D
from astropy.units import Quantity
from astropy.wcs import WCS
from numpy import array

def copy_header_keywords(in_header: Header, out_header: Header):
    for key in in_header.keys():
        if key not in out_header.keys():
            out_header[key] = in_header[key]

def make_cutout_2D(
    image: str,
    pos: SkyCoord,
    size: Quantity,
    outfile: str,
):
    """Make a square cutout of given size using Astropy's Cutout2D function.

    Args:
        image: name of the FITS image to make a cutout of.
        pos: position on which to centre the cutout.
        size: width of the cutout.
        outfile: name of the trimmed output image.
    """
    head = fits.getheader(image)
    data = fits.getdata(image).squeeze()
    if len(data.shape) > 2:
        raise ValueError(
            "Image data has dimensions other than RA and DEC, which is not yet supported."
        )
    wcs = WCS(head).celestial

    cutout = Cutout2D(data, pos, size, wcs=wcs)
    head_new = cutout.wcs.to_header()
    copy_header_keywords(head, head_new)
    hdu = fits.PrimaryHDU(data=cutout.data, header=head_new)
    hdu.writeto(outfile)


def make_cutout_2D_fast(
    image: str,
    pos: SkyCoord,
    size: Quantity,
    outfile: str,
):
    """Make a square cutout of given size by slicing instead of Astropy's Cutout2D function.

    Args:
        image: name of the FITS image to make a cutout of.
        pos: position on which to centre the cutout.
        size: width of the cutout.
        outfile: name of the trimmed output image.
    """
    with fits.open(image) as hdul:
        hdu = hdul[0]
        wcs = WCS(hdu.header).celestial

        pix_pos = wcs.world_to_pixel(pos)

        width_pix = np.floor(size.to("deg").value / abs(hdu.header["CDELT1"]))
        height_pix = np.floor(size.to("deg").value / abs(hdu.header["CDELT2"]))

        # Calculate the bounding box in pixel coordinates
        x_min = int(np.floor(pix_pos[0]) + 1 - width_pix // 2)
        x_max = int(np.floor(pix_pos[0]) + 1 + width_pix // 2 + 1)
        y_min = int(np.floor(pix_pos[1]) + 1 - height_pix // 2)
        y_max = int(np.floor(pix_pos[1]) + 1 + height_pix // 2 + 1)
        print(x_min, x_max, y_min, y_max)

        cutout_data = hdu.data[..., y_min:y_max, x_min:x_max]

        hdu.header["NAXIS1"] = width_pix
        hdu.header["NAXIS2"] = height_pix
        hdu.header["CRPIX1"] -= x_min
        hdu.header["CRPIX2"] -= y_min

        hdu_out = fits.PrimaryHDU(data=cutout_data, header=hdu.header)
        hdu_out.writeto(outfile)


def make_cutout_region(image: str, region: str, outfile: str):
    """Make a cutout from a FITS image using a DS9 region file.

    Args:
        image: name of the FITS image to make a cutout of.
        region: the region to which the image will be trimmed, in DS9 format.
        outfile: name of the trimmed output image.

    Raises:
        NotImplementedError: for cubes with 3 or more dimensions.
        RuntimeError: if the region cannot be parsed for whatever reason.
    """
    hdu = fits.open(image)
    head = hdu[0].header
    data = hdu[0].data.squeeze()

    if data.ndim > 2:
        raise NotImplementedError("Region cutouts for >2D images not supported.")

    r = pyregion.open(region).as_imagecoord(header=head)
    mask = r.get_mask(hdu=hdu[0], shape=(head["NAXIS1"], head["NAXIS2"]))

    assert len(r) == 1, "Multiple regions in 1 file given, only one allowed"

    shape = array(r[0].coord_list)

    # circle
    if len(shape) == 3:
        cutout = Cutout2D(
            data=data * mask,
            position=(shape[0], shape[1]),
            size=(shape[2], shape[2]),
            wcs=WCS(head, naxis=2),
            mode="partial",
        )
    # square
    elif len(shape) > 3:
        print(shape)

        cutout = Cutout2D(
            data=data * mask,
            position=(shape[0], shape[1]),
            size=(shape[3], shape[2]),
            wcs=WCS(head, naxis=2),
            mode="partial",
        )
    else:
        raise RuntimeError("Non-square or circular region found.")

    head_new = cutout.wcs.to_header()
    copy_header_keywords(head, head_new)
    hdu = fits.PrimaryHDU(data=cutout.data, header=head_new)
    hdu.writeto(outfile)



import glob
import logging
import os
import subprocess
import sys
from collections.abc import Iterable
from urllib.request import urlopen

import astropy.units as u
import numpy as np
import pyvo
import requests
import tqdm
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import skycoord_to_pixel
from astroquery.hips2fits import hips2fits
from astroquery.skyview import SkyView

logging.basicConfig(
    format="[%(name)s] %(asctime)s - %(levelname)s: %(message)s", level=logging.INFO
)
logger = logging.getLogger("EveryStamp:Downloader")


def flatten(xs: list) -> Iterable:
    """Flatten a nested list, as per the example on https://stackoverflow.com/a/40857703.

    Args:
        xs : list
            Nested list to flatten.

    Returns:
        x : list
            Flattened list.
    """
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


class FileDownloader(object):
    """From https://medium.com/better-programming/python-progress-bars-with-tqdm-by-example-ce98dbbc9697
    Copyright 2019 tiptapcode Authors. All Rights Reserved.
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
         http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    """

    def get_url_filename(self, url: str) -> str:
        """ Discover file name from HTTP URL, If none is discovered derive name from http redirect HTTP content header Location

        Args:
            url (str): URL from which to get the file name.

        Returns:
            filename (str): name of the file that will be downloaded.

        Raises:
            requests.exceptions.HTTPError: if an HTTP error code is encountered.
            requests.exceptions.ConnectionError: if the connection fails to succeed.
            requests.exceptions.Timeout: if the connection times out.
            requests.exceptions.RequestException: if an error different than above occurs.
        """
        try:
            filename = os.path.basename(url)
            basename, ext = os.path.splitext(filename)
            if ext:
                return filename
            header = requests.head(url, allow_redirects=False).headers
            return (
                os.path.basename(header.get("Location"))
                if "Location" in header
                else filename
            )
        except requests.exceptions.HTTPError as errh:
            print("Http Error:", errh)
            raise errh
        except requests.exceptions.ConnectionError as errc:
            print("Error Connecting:", errc)
            raise errc
        except requests.exceptions.Timeout as errt:
            print("Timeout Error:", errt)
            raise errt
        except requests.exceptions.RequestException as err:
            print("Oops: Something went wrong", err)
            raise err

    def download_file(self, url: str, filename=None, target_dir=None, credentials=None) -> str:
        """ Stream downloads files via HTTP

        Args:
            url (str): URL to download file from.
            filename (str): filename to download to. Overrides filename defined in the URL.
            target_dir (str): directory to download file to.
            credentials (tuple): login credentials if needed.

        Returns:
            target_dest_dir (str): directory where the file will be downloaded to.

        Raises:
            ValueError: 
        """
        if target_dir and not os.path.isdir(target_dir):
            raise ValueError("Invalid target_dir={} specified".format(target_dir))
        local_filename = self.get_url_filename(url) if not filename else filename

        if credentials:
            req = requests.get(url, stream=True, verify=True, auth=credentials)
        else:
            req = requests.get(url, stream=True, verify=True)
        req.raise_for_status()
        try:
            file_size = int(req.headers["Content-Length"])
        except KeyError:
            if req.headers["Transfer-Encoding"] == "chunked":
                file_size = 0
        chunk_size = 1024  # 1 MB
        num_bars = int(file_size / chunk_size)

        base_path = os.path.abspath(os.path.dirname(__file__))
        target_dest_dir = (
            os.path.join(base_path, local_filename)
            if not target_dir
            else os.path.join(target_dir, local_filename)
        )
        with open(target_dest_dir, "wb") as fp:
            for chunk in tqdm.tqdm(
                req.iter_content(chunk_size=chunk_size),
                total=num_bars,
                unit="KB",
                leave=True,
                file=sys.stdout,
            ):
                fp.write(chunk)

        return target_dest_dir


class LoTSSDownloader(FileDownloader):
    """Downloader sub-class for the LOFAR Two-metre Sky Survey."""

    supported_keywords = ["ra", "dec", "size_arcmin"]
    logger = logging.getLogger("EveryStamp:LoTSSDownloader")

    def __init__(self):
        self.url = "https://lofar-surveys.org/{release:s}-cutout.fits?pos={coord_str:s}&size={size_arcmin:f}"

    def format_url(self, ra: str, dec: str, release: str, size: float = 1.0) -> str:
        """

        Args:
            ra: right ascension in HH:MM:SS format.
            dec: declination in HH:MM:SS format.
            size: size of the cutout area in degrees.

        Returns:
            url: the formatted URL.
        """
        size_arcmin = size * 60
        coord_str = (
            SkyCoord(ra, dec, unit="deg", frame="icrs")
            .to_string("hmsdms")
            .replace("h", ":")
            .replace("m", ":")
            .replace("d", ":")
            .replace("s", "")
        )
        url = self.url.format(
            coord_str=coord_str, size_arcmin=size_arcmin, release=release
        )
        return url

    def download(self, **kwargs):
        if kwargs["mode"] != "fits":
            raise ValueError("LoTSSDownloader only supports FITS downloads.")
        furl = self.format_url(
            ra=kwargs["ra"],
            dec=kwargs["dec"],
            size=kwargs["size"],
            release=kwargs["release"],
        )
        logger.info(furl)
        fname = "LoTSS-{release:s}_{ra:f}_{dec:f}_{size:.3f}.{mode:s}".format(
            ra=kwargs["ra"],
            dec=kwargs["dec"],
            mode="fits",
            size=kwargs["size"],
            release=kwargs["release"].upper(),
        )
        if not kwargs["ddir"]:
            self.logger.info(
                "Download directory not specified, downloading to %s instead",
                os.getcwd(),
            )
            ddir = os.getcwd()
        else:
            ddir = kwargs["ddir"]
        if "dr3" in kwargs["release"]:
            self.download_file(furl, filename=fname, target_dir=ddir, credentials=kwargs["credentials"])
        else:
            self.download_file(furl, filename=fname, target_dir=ddir)

class FIRSTDownloader(FileDownloader):
    """Downloader sub-class for the FIRST survey."""

    supported_keywords = ["ra", "dec", "mode"]
    logger = logging.getLogger("EveryStamp:LegacyDownloader")

    def __init__(self):
        self.url_fits = "https://third.ucllnl.org/cgi-bin/firstcutout?RA={ra:s}&Dec={dec:s}&Equinox=J2000&ImageSize={size:f}&ImageType=FITS+File&MaxInt=10&Epochs=&Fieldname=&.submit=+Extract+the+Cutout+&.cgifields=ImageType"

        self.url_image = "https://third.ucllnl.org/cgi-bin/firstimage?RA={ra:s}&Dec={dec:s}&Equinox=J2000&ImageSize={size:f}&MaxInt=10"

    def format_url(
        self,
        ra=None,
        dec=None,
        size=None,
        mode="jpeg",
        **kwargs,
    ):
        """Returns a properly formatted URL that can be used to obtain a cutout from Legacy."""
        pos = SkyCoord(ra, dec, frame="icrs", unit="deg")
        ra = pos.ra.to_string("h", sep=":")
        dec = pos.dec.to_string(sep=":")
        size *= 60
        if mode == "jpeg":
            return self.url_image.format(
                ra=ra,
                dec=dec,
                size=size,
            )
        elif mode == "fits":
            return self.url_fits.format(
                ra=ra,
                dec=dec,
                size=size,
            )

    def download(self, **kwargs):
        if kwargs["mode"] == "both":
            raise NotImplementedError("Download mode `both` is not supported for FIRSTDownloader.")
        else:
            furl = self.format_url(**kwargs)
            self.logger.info("Downloading cutout from %s", furl)
            if not kwargs["ddir"]:
                self.logger.info(
                    "Download directory not specified, downloading to %s instead",
                    os.getcwd(),
                )
                ddir = os.getcwd()
            else:
                ddir = kwargs["ddir"]
            fname = "FIRST_{ra:f}_{dec:f}_{size:.3f}.{mode:s}".format(
                ra=kwargs["ra"],
                dec=kwargs["dec"],
                size=kwargs["size"],
                mode=kwargs["mode"],
            )
        try:
            self.download_file(furl, filename=fname, target_dir=ddir)
        except requests.exceptions.HTTPError:
            self.logger.warning(f"Failed to download {fname}")

class LegacyDownloader(FileDownloader):
    """Downloader sub-class for the DESI Legacy Imaging Surveys."""

    supported_keywords = ["ra", "dec", "mode", "layer", "pixscale", "bands", "size_pix"]
    logger = logging.getLogger("EveryStamp:LegacyDownloader")

    def __init__(self):
        self.url = "https://www.legacysurvey.org/viewer/{mode:s}-cutout/?ra={ra:f}&dec={dec:f}&layer={layer:s}&pixscale={pixscale:.3f}&bands={bands:s}&size={size_pix:d}"

    def format_url(
        self,
        ra=None,
        dec=None,
        size=None,
        bands="grz",
        mode="jpeg",
        layer="ls-dr9",
        pixscale=0.262,
        autoscale=False,
        **kwargs,
    ):
        """Returns a properly formatted URL that can be used to obtain a cutout from Legacy."""
        size_pix = int(size * 3600 / pixscale)
        dlpixscale = pixscale
        dlsize_pix = size_pix
        if (size_pix > 3000) and autoscale:
            # Jump to the next available pixel size by scaling from the (approximate) native pixel scale.
            new_pixscale = pixscale
            new_size_pix = int(size * 3600 / new_pixscale)
            while new_size_pix > 3000:
                new_pixscale += 0.262
                new_size_pix = int(size * 3600 / new_pixscale)
            self.logger.warn(
                "Image size of {:.2f} deg with pixel scale {:.3f} exceeds server limit of 3000 pixels! Automatically adjusting pixel scale to {:.3f} giving {:d} pixels.".format(
                    size, pixscale, new_pixscale, new_size_pix
                ),
                stacklevel=2,
            )
            dlpixscale = new_pixscale
            dlsize_pix = new_size_pix
        elif size_pix > 3000:
            self.logger.warn(
                "Image size of {:.2f} deg with pixel scale {:.3f} exceeds server limit of 3000 pixels! Image will be truncated! Use --legacy_autoscale or pass autoscale=True to automatically switch pixel scales.".format(
                    size, pixscale
                ),
                stacklevel=2,
            )
        return self.url.format(
            ra=ra,
            dec=dec,
            size_pix=dlsize_pix,
            bands=bands,
            mode=mode,
            layer=layer,
            pixscale=dlpixscale,
        )

    def download(self, **kwargs):
        if kwargs["mode"] == "both":
            furl = self.format_url(**kwargs)
            self.logger.info("Downloading cutout from %s", furl)
            if not kwargs["ddir"]:
                self.logger.info(
                    "Download directory not specified, downloading to %s instead",
                    os.getcwd(),
                )
                ddir = os.getcwd()
            else:
                ddir = kwargs["ddir"]
            fname = "legacystamps_{ra:f}_{dec:f}_{layer:s}.{mode:s}".format(
                ra=kwargs["ra"],
                dec=kwargs["dec"],
                layer=kwargs["layer"],
                mode="jpeg",
            )
            self.download_file(furl, filename=fname, target_dir=ddir)

            # Download FITS
            self.logger.info("Downloading cutout from %s", furl)
            furl = furl.replace("jpeg", "fits")
            fname = fname.replace("jpeg", "fits")
            self.download_file(furl, filename=fname, target_dir=ddir)
        else:
            furl = self.format_url(**kwargs)
            self.logger.info("Downloading cutout from %s", furl)
            if not kwargs["ddir"]:
                self.logger.info(
                    "Download directory not specified, downloading to %s instead",
                    os.getcwd(),
                )
                ddir = os.getcwd()
            else:
                ddir = kwargs["ddir"]
            fname = "legacystamps_{ra:f}_{dec:f}_{layer:s}.{mode:s}".format(
                ra=kwargs["ra"],
                dec=kwargs["dec"],
                layer=kwargs["layer"],
                mode=kwargs["mode"],
            )
        try:
            self.download_file(furl, filename=fname, target_dir=ddir)
        except requests.exceptions.HTTPError:
            self.logger.warning(f"Failed to download {fname}")


class PanSTARRSDownloader:
    """Downloader sub-class for the PanSTARRS survey."""

    from panstamps.downloader import downloader as psdownloader

    def __init__(self):
        self.logger = logging.getLogger("EveryStamp:Pan-STARRSDownloader")

    def download(self, ra, dec, size, mode="jpeg", ddir="", bands="gri"):
        if mode == "jpeg":
            get_jpeg = True
            get_fits = False
        elif mode == "fits":
            get_jpeg = False
            get_fits = True
        elif mode == "both":
            get_jpeg = True
            get_fits = True
        arcsecsize = size * 3600
        self.logger.info("Downloading cutout from PANSTARRS")
        d = self.psdownloader(
            ra=ra,
            dec=dec,
            downloadDirectory=ddir or os.getcwd(),
            fits=get_fits,
            jpeg=get_jpeg,
            color=True,
            singleFilters=True,
            filterSet=bands,
            imageType="stack",
            arcsecSize=arcsecsize,
            log=self.logger,
        )
        fitspath, jpegpath, colorpaths = d.get()
        # Rename the output slightly so the user can find it easier.
        for colorpath in colorpaths:
            os.rename(
                colorpath,
                os.path.join(
                    os.path.dirname(colorpath),
                    "panstamps_" + os.path.basename(colorpath),
                ),
            )


class VLASSDownloader(FileDownloader):
    """Downloader sub-class for the VLASS survey.

    Based on the original code by Anna Ho (https://github.com/annayqho/Query_VLASS) and edits by R. Timmerman.
    """

    def __init__(self, datatype: str = "ql"):
        self.summary_url = "https://archive-new.nrao.edu/vlass/VLASS_dyn_summary.php"
        self.logger = logging.getLogger("EveryStamp:VLASSDownloader")
        self.datatype = datatype.lower()
        if self.datatype != "ql" and self.datatype != "se":
            raise ValueError(
                "Unknown VLASS image type specificed. Choose either ql or se for Quick Look or Single Epoch."
            )
        if self.datatype == "ql":
            self.pixel_scale = 2.777777777778e-4 * 3600  # arcsec / pixel
        elif self.datatype == "se":
            self.pixel_scale = 1.666666666667e-04 * 3600

    def get_tiles(self, summary_file="VLASS_dyn_summary.php"):
        """
        Read tiles from tile catalog. If file missing, try wget https://archive-new.nrao.edu/vlass/VLASS_dyn_summary.php

        Parameters
        ----------
        summary_file : str
            Location where the VLASS summary file is located.

        Returns
        -------
        tab : astropy Table
            Table containing tile information.
        """
        if not os.path.isfile(summary_file):
            self.logger.warn(f"Could not find VLASS summary file {summary_file}!")
            self.logger.info("Attempting to download VLASS summary file")
            subprocess.run(
                [
                    "wget",
                    "https://archive-new.nrao.edu/vlass/VLASS_dyn_summary.php",
                    "-O",
                    summary_file,
                ]
            )

        # Put it in a more managable format by replacing consecutive white space with commas.
        # Assumes no more than 1 space in valid entries.
        subprocess.run(["sed", "-i", "-e", r"s/ \{2,\}/,/g", summary_file], check=True)

        tab = Table.read(summary_file, data_start=3, format="ascii.csv")
        tab.rename_column("Observing", "Epoch")
        tab.rename_column("Observation", "Date")
        return tab

    def search_tiles(self, tiles, c):
        """Search the tile catalog for tiles containing the input coordinate

        Parameters
        ----------
        tiles : astropy Table
            Tile catalogue as obtained from get_tiles()
        c : SkyCoord
            Location to check for coverage in one of the observed tiles.

        Returns
        -------
        tile name : str
            Best matching tile.
        observing epoch : str
            Epoch the best matching tile was observed in.
        observing date : str
            Date the best matching tile was observed at.
        """
        ra_h = c.ra.hour
        dec_d = c.dec.deg

        has_dec = np.logical_and(dec_d >= tiles["Dec min"], dec_d < tiles["Dec max"])
        has_ra = np.logical_and(ra_h >= tiles["RA min"], ra_h < tiles["RA max"])
        in_tile = np.logical_and(has_ra, has_dec)
        name = tiles["Tile"][in_tile]
        epoch = tiles["Epoch"][in_tile]
        date = tiles["Date"][in_tile]
        if len(name) == 0:
            raise IndexError("Zero VLASS tiles available for the given coordinate")
        c_grid = SkyCoord(
            7.5 * (tiles["RA min"][in_tile] + tiles["RA max"][in_tile]),
            0.5 * (tiles["Dec min"][in_tile] + tiles["Dec max"][in_tile]),
            unit="deg",
            frame="icrs",
        )
        dist = c_grid.separation(c)
        best_idx = np.argmin(dist)
        return name[best_idx], epoch[best_idx], date[best_idx]

    def get_subtiles(self, tilename, epoch, consider_QA_rejected):
        """For a given tile name, get the subtile filenames in the VLASS directory

        Parse those filenames and return a list of subtile RA and Dec.
        RA and Dec returned as a SkyCoord object

        Parameters
        ----------
        tilename : str
            Name of the tile to extract a subtile from.
        epoch : str
            The epoch the tile was observed in.
        consider_QA_rejected : bool
            Also consider tiles that did not pass the quality assurance checks.

        Returns
        -------
        fname : str
            Name of the subtile.
        c : astropy SkyCoord
            Coordinate of the subtile.
        """

        # Obtain the HTML for the given tile
        if "1." not in epoch:
            if self.datatype == "se":
                URLPATH = f"https://archive-new.nrao.edu/vlass/se_continuum_imaging/{epoch}/{tilename}"
            elif self.datatype == "ql":
                URLPATH = (
                    f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}/{tilename}"
                )
            self.logger.info(f"Downloading from {URLPATH}")
            urlpath = urlopen(f"{URLPATH}")
        else:
            self.logger.info(
                f"Downloading from https://archive-new.nrao.edu/vlass/quicklook/{epoch}v2/{tilename}"
            )
            urlpath = urlopen(
                f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}v2/{tilename}"
            )
        string = urlpath.read().decode("utf-8").split("\n")

        if self.datatype == "ql" and consider_QA_rejected:
            # Obtain the HTML for the QA Rejected
            urlpath_rejected = urlopen(
                f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}v2/QA_REJECTED"
            )
            string += urlpath_rejected.read().decode("utf-8").split("\n")

        # Select only the subtile parts
        vals = np.array(
            [
                val.strip()
                for val in string
                if ("href" in val.strip()) and (tilename in val.strip())
            ]
        )

        # Select the coordinate part. You want the 'VLASS1.1.ql.T25t12.J150000+603000.10.2048.v1/' bit
        fname = np.array([val.split('"')[7] for val in vals])

        # Split out the actual coordinate string
        pos_raw = np.array(
            [val.split(".")[4] for val in fname if val.startswith("VLASS")]
        )

        if "-" in pos_raw[0]:
            # dec < 0
            ra_raw = np.array([val.split("-")[0] for val in pos_raw])
            dec_raw = np.array([val.split("-")[1] for val in pos_raw])
        else:
            # dec > 0
            ra_raw = np.array([val.split("+")[0] for val in pos_raw])
            dec_raw = np.array([val.split("+")[1] for val in pos_raw])
        ra = []
        dec = []
        for ii, val in enumerate(ra_raw):
            if val[1:3] == "24":
                rah = "00"
            else:
                rah = val[1:3]
            ra.append(f"{rah}h{val[3:5]}m{val[5:]}s")
            dec.append(f"{dec_raw[ii][:2]}d{dec_raw[ii][2:4]}m{dec_raw[ii][4:]}s")
        ra = np.array(ra)
        dec = np.array(dec)
        c = SkyCoord(
            ra, dec, frame="icrs"
        )  # .directional_offset_by(45*u.deg, 0.75*u.deg)
        return fname, c

    def get_cutout(self, imname, c, crop_scale):
        """Get a smaller cutout from the subtile.

        Parameters
        ----------
        imname : str
            Name of the image to make a cutout from.
        c : astropy SkyCoord
            Coordinate around which to make a cutout.
        crop_scale : int
            Size of the cutout in pixels.

        Returns
        -------
        output_fits : str
            Name of the output FITS file.
        """
        # Define output name
        output_fits = os.path.join(
            "/".join(imname.split("/")[:-1]),
            "VLASS_{:.6f}_{:.6f}_".format(c.ra.value, c.dec.value)
            + imname.split("/")[-1].rstrip(".fits")
            + "_poststamp.fits",
        )

        # Get header info
        hdu_list = fits.open(imname)
        header = hdu_list[0].header
        data = hdu_list[0].data[0, 0, :, :]

        # Obtain header and drop useless axes
        wcs = WCS(header)
        wcs = wcs.dropaxis(2).dropaxis(2)

        pixel_coords = skycoord_to_pixel(
            SkyCoord(c.ra.deg, c.dec.deg, unit="deg", frame="icrs"), wcs
        )

        if (
            pixel_coords[0] < 0
            or pixel_coords[1] < 0
            or pixel_coords[0] > data.shape[0]
            or pixel_coords[1] > data.shape[1]
        ):
            subprocess.call(f"rm -f {imname}", shell=True)
            raise Exception(
                "Requested coordinate not within the available subtiles. Consider running with consider_QA_rejected=True to also search additional subtiles which failed initial QA checks"
            )

        # Produce a cutout
        cutout = Cutout2D(data, c, (crop_scale, crop_scale), wcs=wcs)

        # Update the HDU
        hdu_list[0].data = cutout.data
        new_header = cutout.wcs.to_header()
        hdu_list[0].header.update(new_header)
        hdu_list[0].header.set("NAXIS", 4)
        hdu_list[0].header.insert("NAXIS2", ("NAXIS3", 1), after=True)
        hdu_list[0].header.insert("NAXIS3", ("NAXIS4", 1), after=True)
        hdu_list[0].header.remove("WCSAXES", ignore_missing=True)
        hdu_list[0].header.remove("MJDREF", ignore_missing=True)
        hdu_list[0].header.remove("MJD-OBS", ignore_missing=True)

        # Write the new fits
        hdu_list.writeto(output_fits, overwrite=True)

        # Cleanup
        subprocess.call(f"rm -f {imname}", shell=True)

        return output_fits

    def search_vlass(
        self,
        c,
        crop=False,
        crop_scale=256,
        consider_QA_rejected=False,
        ddir=os.getcwd(),
    ):
        """
        Searches the VLASS catalog for a source

        Parameters
        ----------
        c : astropy SkyCoord
            Coordinate to search for in tiles.
        crop : bool
            Make a cropped cutout of the area of interest.
        crop_scale : int
            Crop the cutout to this amount of pixels centred around c.
        consider_QA_rejected : bool
            Also consider tiles that failed the Quality Assurance checks.
        ddir : str
            Location to download the cutout to.

        Returns
        -------
        imname : str
            Name of the output image.
        """
        # Find the VLASS tile
        tiles = self.get_tiles()
        tilename, epoch, obsdate = self.search_tiles(tiles, c)
        if self.datatype == "se":
            # No other epoch available for now, so hardcode until the code is updated.
            epoch = "VLASS2.1"

        subtiles, c_tiles = self.get_subtiles(tilename, epoch, consider_QA_rejected)
        dist = c.separation(c_tiles)
        subtile = subtiles[np.argmin(dist)]

        imname = f"{subtile[:-1]}.I.iter1.image.pbcor.tt0.subim.fits"
        if len(glob.glob(imname)) == 0:
            if self.datatype == "ql":
                if "1." not in epoch:
                    url_get = f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}/{tilename}/{subtile}"
                else:
                    url_get = f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}v2/{tilename}/{subtile}"
            elif self.datatype == "se":
                # Dirty workaround until proper implementation.
                epoch = "VLASS2.1"
                imname = imname.replace("iter1", "iter3")
                url_get = f"https://archive-new.nrao.edu/vlass/se_continuum_imaging/{epoch}/{tilename}/{subtile}"
            fname = f"{url_get}{imname}"
            self.logger.info("Downloading to " + ddir)
            self.download_file(fname, target_dir=ddir)
            if self.datatype == "ql" and consider_QA_rejected:
                if "1." not in epoch:
                    url_get = f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}/QA_REJECTED/{subtile}"
                else:
                    url_get = f"https://archive-new.nrao.edu/vlass/quicklook/{epoch}v2/QA_REJECTED/{subtile}"
                fname = f"{url_get}{imname}"
                self.download_file(fname, target_dir=ddir)
        if crop:
            out = self.get_cutout(os.path.join(ddir, imname), c, crop_scale=crop_scale)
            return out
        else:
            return imname

    def download(
        self,
        ra=0.0,
        dec=0.0,
        size=0.1,
        crop=True,
        consider_QA_rejected=False,
        ddir=os.getcwd(),
    ):
        """Download a cutout from the VLASS survey.

        Parameters
        ----------
        ra : float
            Right ascension of the coordinate of interest.
        dec : float
            Declination of the coordinate of interest.
        size : float
            Size of the area of interest in degrees.
        crop : bool
            Crop the image to the area of interest.
        consider_QA_rejected : bool
            Also consider tiles that failed the Quality Assurance checks.
        ddir : str
            Location to download the cutout to.
        """
        c = SkyCoord(ra, dec, unit="deg")

        crop_scale = size * 3600 / self.pixel_scale
        self.logger.info("Downloading cutout from VLASS")
        self.search_vlass(
            c,
            crop=crop,
            crop_scale=crop_scale,
            consider_QA_rejected=consider_QA_rejected,
            ddir=ddir,
        )


class VODownloader:
    """Downloader sub-class for surveys offeret through a VO."""

    def __init__(self, url, name=""):
        if not url:
            raise ValueError("VO url cannot be empty.")
        else:
            self.url = url
        if not name:
            self.name = self.url
            self.logger = logging.getLogger("EveryStamp:VODownloader")
        else:
            self.name = name
            self.logger = logging.getLogger(
                "EveryStamp:VODownloader[{:s}]".format(self.name)
            )

    def download(self, ra=0.0, dec=0.0, size=0.1, ddir=os.getcwd(), suffix=""):
        """Download a cutout from a VO survey.

        Parameters
        ----------
        ra : float
            Right ascension of the coordinate of interest in degrees.
        dec : float
            Declination of the coordinate of interest in degrees    .
        size : float
            Size of the area of interest in degrees.
        ddir : str
            Location to download the cutout to.
        """
        c = SkyCoord(ra, dec, unit="deg")

        vo = pyvo.sia.SIAService(self.url)
        query = vo.search(c, size=size)
        if not query:
            raise ValueError("Requested coordinates not covered by the specified VO!")
        im = query.getrecord(0)
        if im.format == "image/fits":
            self.logger.info("Downloading cutout from {:s}".format(self.name))
            im.cachedataset(
                filename=os.path.join(
                    ddir,
                    "{:s}_{:.4f}_{:.4f}_{:.3f}.fits".format(self.name, ra, dec, size),
                )
            )


class HiPSDownloader:
    """Sub-class to download a file from a HiPS image using hips2fits."""

    def __init__(self, hips, name=""):
        if not hips:
            raise ValueError("HiPS name cannot be empty.")
        else:
            self.hips = hips
        if not name:
            self.name = self.hips.replace("/", "_")
        else:
            self.name = name
        self.logger = logging.getLogger(
            "EveryStamp:HiPSDownloader[{:s}]".format(self.name)
        )
        self.logger.warning(
            "downloading from a HiPS survey. Scientific accuracy of resulting FITS images is not guaranteed!"
        )

    def download(
        self, ra=0.0, dec=0.0, size=0.1, ddir=os.getcwd(), pixsize=1.0, mode="jpg"
    ):
        """Download a cutout from a HiPS survey.

        Parameters
        ----------
        ra : float
            Right ascension of the coordinate of interest in degrees.
        dec : float
            Declination of the coordinate of interest in degrees    .
        size : float
            Size of the area of interest in degrees.
        ddir : str
            Location to download the cutout to.
        pixsize : float
            Pixel scale of the survey in arcsec. Default is 1.0 arcsec per pixel.
        mode : str
            What image format to download. Can be jpg or fits, default is jpg.
        """
        imsize = int(size / (pixsize / 3600))
        img = hips2fits.query(
            hips=self.hips,
            format=mode,
            width=imsize,
            height=imsize,
            projection="SIN",
            fov=size * u.deg,
            ra=ra * u.deg,
            dec=dec * u.deg,
        )
        if not os.path.exists(ddir):
            self.logger.info("Download directory does not exist, creating it")
            os.mkdir(ddir)
        if mode == "jpg":
            from PIL import Image

            imdata = Image.fromarray(img)
            imdata.save(
                os.path.join(
                    ddir,
                    "{:s}_{:.4f}_{:.4f}_{:.5f}.jpeg".format(self.name, ra, dec, size),
                )
            )
        elif mode == "fits":
            img.writeto(
                os.path.join(
                    ddir,
                    "{:s}_{:.4f}_{:.4f}_{:.5f}.fits".format(self.name, ra, dec, size),
                )
            )


class SkyViewDownloader:
    """Downloader sub-class for surveys offeret through a VO."""

    def __init__(self, survey):
        if not survey:
            raise ValueError("SkyView survey cannot be empty.")
        else:
            self.survey = survey
            self.logger = logging.getLogger(
                "EveryStamp:SkyViewDownloader[{:s}]".format(self.survey)
            )

    def download(
        self, ra=0.0, dec=0.0, size=0.1, pixsize=1, ddir=os.getcwd(), suffix=""
    ):
        """Download a cutout from a SkyView survey.

        Parameters
        ----------
        ra : float
            Right ascension of the coordinate of interest in degrees.
        dec : float
            Declination of the coordinate of interest in degrees    .
        size : float
            Size of the area of interest in degrees.
        pixsize
            Pixel size to use for the cutout.
        ddir : str
            Location to download the cutout to.
        """
        c = SkyCoord(ra, dec, unit="deg")

        sv = SkyView()
        pixsize_deg = pixsize / 3600.0
        pixels = size / pixsize_deg
        self.logger.info(
            f"Requesting a {size} x {size} deg cutout of {pixels} x {pixels} pixels."
        )
        hdul = sv.get_images(c, self.survey, radius=size * u.deg, pixels=int(pixels))
        if not hdul:
            raise ValueError(
                "SkyView did not return a result. If you requested a large cutout, try increasing the pixel size through --skyview_pixsize or reducing the cutout area."
            )
        hdul[0].writeto(
            os.path.join(
                ddir,
                "{:s}_{:.4f}_{:.4f}_{:.3f}.fits".format(
                    self.survey.replace(" ", "_"), ra, dec, size
                ),
            )
        )
